from __future__ import annotations

from typing import Any

from smac.intensifier.successive_halving import SuccessiveHalving


class Hyperband(SuccessiveHalving):
    """See ``SuccessiveHalving`` for documentation."""

    def reset(self) -> None:
        """Resets the internal variables of the intensifier, including the tracker and the next bracket."""
        super().reset()

        # Reset current bracket
        self._next_bracket: int = 0

    def __post_init__(self) -> None:
        super().__post_init__()

        min_budget = self._min_budget
        max_budget = self._max_budget
        assert min_budget is not None and max_budget is not None
        eta = self._eta

        # The only difference we have to do is change max_iterations, n_configs_in_stage, budgets_in_stage
        self._s_max = self._get_max_iterations(eta, max_budget, min_budget)  # type: ignore[operator]
        self._max_iterations: dict[int, int] = {}
        self._n_configs_in_stage: dict[int, list] = {}
        self._budgets_in_stage: dict[int, list] = {}

        for i in range(self._s_max + 1):
            max_iter = self._s_max - i

            self._budgets_in_stage[i], self._n_configs_in_stage[i] = self._compute_configs_and_budgets_for_stages(
                eta, max_budget, max_iter, self._s_max
            )
            self._max_iterations[i] = max_iter + 1

    def get_state(self) -> dict[str, Any]:  # noqa: D102
        state = super().get_state()
        state["next_bracket"] = self._next_bracket

        return state

    def set_state(self, state: dict[str, Any]) -> None:  # noqa: D102
        super().set_state(state)
        self._next_bracket = state["next_bracket"]

    def _get_next_bracket(self) -> int:
        """In contrast to Successive Halving, Hyperband uses multiple brackets. Each time a new batch
        is added to the tracker, the bracket is increased.
        """
        current_bracket = self._next_bracket
        next_bracket = current_bracket + 1

        if next_bracket > self._s_max or next_bracket < 0:
            next_bracket = 0

        self._next_bracket = next_bracket

        return current_bracket



from __future__ import annotations

from ConfigSpace import Configuration

from smac.acquisition.function.expected_improvement import EI
from smac.acquisition.maximizer.local_and_random_search import (
    LocalAndSortedRandomSearch,
)
from smac.facade.abstract_facade import AbstractFacade
from smac.initial_design.sobol_design import SobolInitialDesign
from smac.intensifier.intensifier import Intensifier
from smac.model.random_forest.random_forest import RandomForest
from smac.multi_objective.aggregation_strategy import MeanAggregationStrategy
from smac.random_design.probability_design import ProbabilityRandomDesign
from smac.runhistory.encoder.log_scaled_encoder import RunHistoryLogScaledEncoder
from smac.scenario import Scenario

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class HyperparameterOptimizationFacade(AbstractFacade):
    @staticmethod
    def get_model(  # type: ignore
        scenario: Scenario,
        *,
        n_trees: int = 10,
        ratio_features: float = 1.0,
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        max_depth: int = 2**20,
        bootstrapping: bool = True,
    ) -> RandomForest:
        """Returns a random forest as surrogate model.

        Parameters
        ----------
        n_trees : int, defaults to 10
            The number of trees in the random forest.
        ratio_features : float, defaults to 5.0 / 6.0
            The ratio of features that are considered for splitting.
        min_samples_split : int, defaults to 3
            The minimum number of data points to perform a split.
        min_samples_leaf : int, defaults to 3
            The minimum number of data points in a leaf.
        max_depth : int, defaults to 20
            The maximum depth of a single tree.
        bootstrapping : bool, defaults to True
            Enables bootstrapping.
        """
        return RandomForest(
            log_y=True,
            n_trees=n_trees,
            bootstrapping=bootstrapping,
            ratio_features=ratio_features,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            max_depth=max_depth,
            configspace=scenario.configspace,
            instance_features=scenario.instance_features,
            seed=scenario.seed,
        )

    @staticmethod
    def get_acquisition_function(  # type: ignore
        scenario: Scenario,
        *,
        xi: float = 0.0,
    ) -> EI:
        """Returns an Expected Improvement acquisition function.

        Parameters
        ----------
        scenario : Scenario
        xi : float, defaults to 0.0
            Controls the balance between exploration and exploitation of the
            acquisition function.
        """
        return EI(xi=xi, log=True)

    @staticmethod
    def get_acquisition_maximizer(  # type: ignore
        scenario: Scenario,
        *,
        challengers: int = 10000,
        local_search_iterations: int = 10,
    ) -> LocalAndSortedRandomSearch:
        """Returns local and sorted random search as acquisition maximizer.

        Warning
        -------
        If you experience RAM issues, try to reduce the number of challengers.

        Parameters
        ----------
        challengers : int, defaults to 10000
            Number of challengers.
        local_search_iterations: int, defaults to 10
            Number of local search iterations.
        """
        optimizer = LocalAndSortedRandomSearch(
            scenario.configspace,
            challengers=challengers,
            local_search_iterations=local_search_iterations,
            seed=scenario.seed,
        )

        return optimizer

    @staticmethod
    def get_intensifier(  # type: ignore
        scenario: Scenario,
        *,
        max_config_calls: int = 3,
        max_incumbents: int = 10,
    ) -> Intensifier:
        """Returns ``Intensifier`` as intensifier. Uses the default configuration for ``race_against``.

        Parameters
        ----------
        scenario : Scenario
        max_config_calls : int, defaults to 3
            Maximum number of configuration evaluations. Basically, how many instance-seed keys should be max evaluated
            for a configuration.
        max_incumbents : int, defaults to 10
            How many incumbents to keep track of in the case of multi-objective.
        """
        return Intensifier(
            scenario=scenario,
            max_config_calls=max_config_calls,
            max_incumbents=max_incumbents,
        )

    @staticmethod
    def get_initial_design(  # type: ignore
        scenario: Scenario,
        *,
        n_configs: int | None = None,
        n_configs_per_hyperparamter: int = 10,
        max_ratio: float = 0.25,
        additional_configs: list[Configuration] | None = None,
    ) -> SobolInitialDesign:
        """Returns a Sobol design instance.

        Parameters
        ----------
        scenario : Scenario
        n_configs : int | None, defaults to None
            Number of initial configurations (disables the arguments ``n_configs_per_hyperparameter``).
        n_configs_per_hyperparameter: int, defaults to 10
            Number of initial configurations per hyperparameter. For example, if my configuration space covers five
            hyperparameters and ``n_configs_per_hyperparameter`` is set to 10, then 50 initial configurations will be
            samples.
        max_ratio: float, defaults to 0.25
            Use at most ``scenario.n_trials`` * ``max_ratio`` number of configurations in the initial design.
            Additional configurations are not affected by this parameter.
        additional_configs: list[Configuration], defaults to []
            Adds additional configurations to the initial design.
        """
        return SobolInitialDesign(
            scenario=scenario,
            n_configs=n_configs,
            n_configs_per_hyperparameter=n_configs_per_hyperparamter,
            max_ratio=max_ratio,
            additional_configs=additional_configs,
        )

    @staticmethod
    def get_random_design(  # type: ignore
        scenario: Scenario,
        *,
        probability: float = 0.2,
    ) -> ProbabilityRandomDesign:
        """Returns ``ProbabilityRandomDesign`` for interleaving configurations.

        Parameters
        ----------
        probability : float, defaults to 0.2
            Probability that a configuration will be drawn at random.
        """
        return ProbabilityRandomDesign(probability=probability, seed=scenario.seed)

    @staticmethod
    def get_multi_objective_algorithm(  # type: ignore
        scenario: Scenario,
        *,
        objective_weights: list[float] | None = None,
    ) -> MeanAggregationStrategy:
        """Returns the mean aggregation strategy for the multi-objective algorithm.

        Parameters
        ----------
        scenario : Scenario
        objective_weights : list[float] | None, defaults to None
            Weights for averaging the objectives in a weighted manner. Must be of the same length as the number of
            objectives.
        """
        return MeanAggregationStrategy(
            scenario=scenario,
            objective_weights=objective_weights,
        )

    @staticmethod
    def get_runhistory_encoder(  # type: ignore
        scenario: Scenario,
    ) -> RunHistoryLogScaledEncoder:
        """Returns a log scaled runhistory encoder. That means that costs are log scaled before
        training the surrogate model.
        """
        return RunHistoryLogScaledEncoder(scenario)



from __future__ import annotations

from typing import Mapping

import numpy as np

from smac.runhistory.encoder import AbstractRunHistoryEncoder
from smac.runhistory.runhistory import TrialKey, TrialValue
from smac.utils.configspace import convert_configurations_to_array
from smac.utils.logging import get_logger
from smac.utils.multi_objective import normalize_costs

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class RunHistoryEncoder(AbstractRunHistoryEncoder):
    def _build_matrix(
        self,
        trials: Mapping[TrialKey, TrialValue],
        store_statistics: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:
        # First build nan-matrix of size #configs x #params+1
        n_rows = len(trials)
        n_cols = self._n_params
        X = np.ones([n_rows, n_cols + self._n_features]) * np.nan

        # For now we keep it as 1
        # TODO: Extend for native multi-objective
        y = np.ones([n_rows, 1])

        # Then populate matrix
        for row, (key, run) in enumerate(trials.items()):
            # Scaling is automatically done in configSpace
            conf = self.runhistory._ids_config[key.config_id]
            conf_vector = convert_configurations_to_array([conf])[0]

            if self._n_features > 0 and self._instance_features is not None:
                assert isinstance(key.instance, str)
                feats = self._instance_features[key.instance]
                X[row, :] = np.hstack((conf_vector, feats))
            else:
                X[row, :] = conf_vector

            if self._n_objectives > 1:
                assert self._multi_objective_algorithm is not None
                assert isinstance(run.cost, list)

                # Let's normalize y here
                # We use the objective_bounds calculated by the runhistory
                y_ = normalize_costs(run.cost, self.runhistory.objective_bounds)
                y_agg = self._multi_objective_algorithm(y_)
                y[row] = y_agg
            else:
                y[row] = run.cost

        if y.size > 0:
            if store_statistics:
                self._percentile = np.percentile(y, self._scale_percentage, axis=0)
                self._min_y = np.min(y, axis=0)
                self._max_y = np.max(y, axis=0)

        y = self.transform_response_values(values=y)
        return X, y

    def transform_response_values(self, values: np.ndarray) -> np.ndarray:
        """Returns the input values."""
        return values



from __future__ import annotations

from typing import Any

import numpy as np

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class LCB(AbstractAcquisitionFunction):
    r"""Computes the lower confidence bound for a given x over the best so far value as acquisition value.

    :math:`LCB(X) = \mu(\mathbf{X}) - \sqrt(\beta_t)\sigma(\mathbf{X})` [[SKKS10][SKKS10]]

    with

    :math:`\beta_t = 2 \log( |D| t^2 / \beta)`

    :math:`\text{Input space} D`
    :math:`\text{Number of input dimensions} |D|`
    :math:`\text{Number of data points} t`
    :math:`\text{Exploration/exploitation tradeoff} \beta`

    Returns -LCB(X) as the acquisition_function optimizer maximizes the acquisition value.

    Parameters
    ----------
    beta : float, defaults to 1.0
        Controls the balance between exploration and exploitation of the acquisition function.

    Attributes
    ----------
    _beta : float
        Exploration-exploitation trade-off parameter.
    _num_data : int
        Number of data points seen so far.
    """

    def __init__(self, beta: float = 1.0) -> None:
        super(LCB, self).__init__()
        self._beta: float = beta
        self._num_data: int | None = None

    @property
    def name(self) -> str:  # noqa: D102
        return "Lower Confidence Bound"

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"beta": self._beta})

        return meta

    def _update(self, **kwargs: Any) -> None:
        """Update acsquisition function attributes

        Parameters
        ----------
        num_data : int
            Number of data points
        """
        assert "num_data" in kwargs
        self._num_data = kwargs["num_data"]

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute LCB acquisition value

        Parameters
        ----------
        X : np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N,1]
            Acquisition function values wrt X.

        Raises
        ------
        ValueError
            If `update` has not been called before. Number of data points is unspecified in this case.
        """
        assert self._model is not None
        if self._num_data is None:
            raise ValueError(
                "No current number of data points specified. Call `update` to inform the acquisition function."
            )

        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        m, var_ = self._model.predict_marginalized(X)
        std = np.sqrt(var_)
        beta_t = 2 * np.log((X.shape[1] * self._num_data**2) / self._beta)

        return -(m - np.sqrt(beta_t) * std)



from __future__ import annotations

from ConfigSpace import Configuration, ConfigurationSpace, Float
from src.datasets.dataset import Dataset
from src.models.model import Model


class HimmelblauModel(Model):
    def __init__(self, dataset: Dataset | None = None) -> None:
        super().__init__(dataset)

    @property
    def configspace(self) -> ConfigurationSpace:
        # Build Configuration Space which defines all parameters and their ranges
        cs = ConfigurationSpace(seed=0)

        # First we create our hyperparameters
        x = Float("x", (-5, 5))
        y = Float("y", (-5, 5))

        # Add hyperparameters and conditions to our configspace
        cs.add([x, y])

        return cs

    def train(self, config: Configuration, seed: int) -> float:
        x = config["x"]
        y = config["y"]

        return (x**2 + y - 11) ** 2 + (x + y**2 - 7) ** 2



from __future__ import annotations

import numpy as np
from ConfigSpace import Configuration

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.maximizer.random_search import RandomSearch
from smac.facade.abstract_facade import AbstractFacade
from smac.initial_design.default_design import DefaultInitialDesign
from smac.intensifier.intensifier import Intensifier
from smac.model.random_model import RandomModel
from smac.multi_objective.aggregation_strategy import MeanAggregationStrategy
from smac.random_design import AbstractRandomDesign
from smac.runhistory.encoder.encoder import RunHistoryEncoder
from smac.scenario import Scenario

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class RandomFacade(AbstractFacade):
    """
    Facade to use Random Online Aggressive Racing (ROAR).

    *Aggressive Racing:*
    When we have a new configuration θ, we want to compare it to the current best
    configuration, the incumbent θ*. ROAR uses the 'racing' approach, where we run few times for unpromising θ and many
    times for promising configurations. Once we are confident enough that θ is better than θ*, we update the
    incumbent θ* ⟵ θ. `Aggressive` means rejecting low-performing configurations very early, often after a single run.
    This together is called `aggressive racing`.

    *ROAR Loop:*
    The main ROAR loop looks as follows:

    1. Select a configuration θ uniformly at random.
    2. Compare θ to incumbent θ* online (one θ at a time):
      - Reject/accept θ with `aggressive racing`

    *Setup:*
    Uses a random model and random search for the optimization of the acquisition function.

    Note
    ----
    The surrogate model and the acquisition function is not used during the optimization and therefore replaced
    by dummies.
    """

    @staticmethod
    def get_acquisition_function(scenario: Scenario) -> AbstractAcquisitionFunction:
        """The random facade is not using an acquisition function. Therefore, we simply return a dummy function."""

        class DummyAcquisitionFunction(AbstractAcquisitionFunction):
            def _compute(self, X: np.ndarray) -> np.ndarray:
                return X

        return DummyAcquisitionFunction()

    @staticmethod
    def get_intensifier(
        scenario: Scenario,
        *,
        max_config_calls: int = 3,
        max_incumbents: int = 10,
    ) -> Intensifier:
        """Returns ``Intensifier`` as intensifier.

        Note
        ----
        Please use the ``HyperbandFacade`` if you want to incorporate budgets.

        Warning
        -------
        If you are in an algorithm configuration setting, consider increasing ``max_config_calls``.

        Parameters
        ----------
        max_config_calls : int, defaults to 3
            Maximum number of configuration evaluations. Basically, how many instance-seed keys should be max evaluated
            for a configuration.
        max_incumbents : int, defaults to 10
            How many incumbents to keep track of in the case of multi-objective.
        """
        return Intensifier(
            scenario=scenario,
            max_config_calls=max_config_calls,
            max_incumbents=max_incumbents,
        )

    @staticmethod
    def get_initial_design(
        scenario: Scenario,
        *,
        additional_configs: list[Configuration] = None,
    ) -> DefaultInitialDesign:
        """Returns an initial design, which returns the default configuration.

        Parameters
        ----------
        additional_configs: list[Configuration], defaults to []
            Adds additional configurations to the initial design.
        """
        if additional_configs is None:
            additional_configs = []
        return DefaultInitialDesign(
            scenario=scenario,
            additional_configs=additional_configs,
        )

    @staticmethod
    def get_random_design(scenario: Scenario) -> AbstractRandomDesign:
        """Just like the acquisition function, we do not use a random design. Therefore, we return a dummy design."""

        class DummyRandomDesign(AbstractRandomDesign):
            def check(self, iteration: int) -> bool:
                return True

        return DummyRandomDesign()

    @staticmethod
    def get_model(scenario: Scenario) -> RandomModel:
        """The model is used in the acquisition function. Since we do not use an acquisition function, we return a
        dummy model (returning random values in this case).
        """
        return RandomModel(
            configspace=scenario.configspace,
            instance_features=scenario.instance_features,
            seed=scenario.seed,
        )

    @staticmethod
    def get_acquisition_maximizer(scenario: Scenario) -> RandomSearch:
        """We return ``RandomSearch`` as maximizer which samples configurations randomly from the configuration
        space and therefore neither uses the acquisition function nor the model.
        """
        return RandomSearch(
            scenario.configspace,
            seed=scenario.seed,
        )

    @staticmethod
    def get_multi_objective_algorithm(  # type: ignore
        scenario: Scenario,
        *,
        objective_weights: list[float] | None = None,
    ) -> MeanAggregationStrategy:
        """Returns the mean aggregation strategy for the multi-objective algorithm.

        Parameters
        ----------
        scenario : Scenario
        objective_weights : list[float] | None, defaults to None
            Weights for averaging the objectives in a weighted manner. Must be of the same length as the number of
            objectives.
        """
        return MeanAggregationStrategy(
            scenario=scenario,
            objective_weights=objective_weights,
        )

    @staticmethod
    def get_runhistory_encoder(scenario: Scenario) -> RunHistoryEncoder:
        """Returns the default runhistory encoder."""
        return RunHistoryEncoder(scenario)



from __future__ import annotations

from typing import Any

import numpy as np
from ConfigSpace import Configuration
from ConfigSpace.hyperparameters import FloatHyperparameter

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.function.confidence_bound import LCB
from smac.acquisition.function.integrated_acquisition_function import (
    IntegratedAcquisitionFunction,
)
from smac.acquisition.function.thompson import TS
from smac.model.abstract_model import AbstractModel
from smac.model.random_forest.abstract_random_forest import AbstractRandomForest
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class PriorAcquisitionFunction(AbstractAcquisitionFunction):
    r"""Weight the acquisition function with a user-defined prior over the optimum.

    See "piBO: Augmenting Acquisition Functions with User Beliefs for Bayesian Optimization" by Carl
    Hvarfner et al. [[HSSL22][HSSL22]] for further details.

    Parameters
    ----------
    decay_beta: float
        Decay factor on the user prior. A solid default value for decay_beta (empirically founded) is
        ``scenario.n_trials`` / 10.
    prior_floor : float, defaults to 1e-12
        Lowest possible value of the prior to ensure non-negativity for all values in the search space.
    discretize : bool, defaults to False
        Whether to discretize (bin) the densities for continous parameters. Triggered for Random Forest models and
        continous hyperparameters to avoid a pathological case where all Random Forest randomness is removed
        (RF surrogates require piecewise constant acquisition functions to be well-behaved).
    discrete_bins_factor : float, defaults to 10.0
        If discretizing, the multiple on the number of allowed bins for each parameter.
    """

    def __init__(
        self,
        acquisition_function: AbstractAcquisitionFunction,
        decay_beta: float,
        prior_floor: float = 1e-12,
        discretize: bool = False,
        discrete_bins_factor: float = 10.0,
    ):
        super().__init__()
        self._acquisition_function: AbstractAcquisitionFunction = acquisition_function
        self._functions: list[AbstractAcquisitionFunction] = []
        self._eta: float | None = None

        self._hyperparameters: dict[Any, Configuration] | None = None
        self._decay_beta = decay_beta
        self._prior_floor = prior_floor
        self._discretize = discretize
        self._discrete_bins_factor = discrete_bins_factor

        # check if the acquisition function is LCB or TS - then the acquisition function values
        # need to be rescaled to assure positiveness & correct magnitude
        if isinstance(self._acquisition_function, IntegratedAcquisitionFunction):
            acquisition_type = self._acquisition_function._acquisition_function
        else:
            acquisition_type = self._acquisition_function

        self._rescale = isinstance(acquisition_type, (LCB, TS))

        # Variables needed to adapt the weighting of the prior
        self._initial_design_size = None
        self._iteration_number = 0

    @property
    def name(self) -> str:  # noqa: D102
        return f"Prior Acquisition Function ({self._acquisition_function.__class__.__name__})"

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "acquisition_function": self._acquisition_function.meta,
                "decay_beta": self._decay_beta,
                "prior_floor": self._prior_floor,
                "discretize": self._discretize,
                "discrete_bins_factor": self._discrete_bins_factor,
            }
        )

        return meta

    @property
    def model(self) -> AbstractModel | None:  # noqa: D102
        return self._model

    @model.setter
    def model(self, model: AbstractModel) -> None:
        self._model = model
        self._hyperparameters = model._configspace.get_hyperparameters_dict()

        if isinstance(model, AbstractRandomForest):
            if not self._discretize:
                logger.warning("Discretizing the prior for random forest models.")
                self._discretize = True

    def _update(self, **kwargs: Any) -> None:
        """Update the acquisition function attributes required for calculation.

        Parameters
        ----------
        eta : float
            Current incumbent value.
        """
        assert "eta" in kwargs

        # Compute intiial design size
        if self._initial_design_size is None:
            self._initial_design_size = kwargs["num_data"]

        self._iteration_number = kwargs["num_data"] - self._initial_design_size
        self._eta = kwargs["eta"]

        assert self.model is not None
        self._acquisition_function.update(model=self.model, **kwargs)

    def _compute_prior(self, X: np.ndarray) -> np.ndarray:
        """Compute the prior-weighted acquisition function values, where the prior on each
        parameter is multiplied by a decay factor controlled by the parameter decay_beta and
        the iteration number. Multivariate priors are not supported, for now.

        Parameters
        ----------
        X: np.ndarray [N, D]
            The input points where the user-specified prior should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N, 1]
            The user prior over the optimum for values of X.
        """
        assert self._hyperparameters is not None

        prior_values = np.ones((len(X), 1))
        # iterate over the hyperparmeters (alphabetically sorted) and the columns, which come
        # in the same order
        for parameter, X_col in zip(self._hyperparameters.values(), X.T):
            if self._discretize and isinstance(parameter, FloatHyperparameter):
                assert self._discrete_bins_factor is not None
                number_of_bins = int(
                    np.ceil(self._discrete_bins_factor * self._decay_beta / (self._iteration_number + 1))
                )
                prior_values *= self._compute_discretized_pdf(parameter, X_col, number_of_bins)
            else:
                prior_values *= parameter._pdf(X_col[:, np.newaxis])

        return prior_values

    def _compute_discretized_pdf(
        self,
        hyperparameter: FloatHyperparameter,
        X_col: np.ndarray,
        number_of_bins: int,
    ) -> np.ndarray:
        """Discretize (bins) prior values on continous a specific continous parameter
        to an increasingly coarse discretization determined by the prior decay parameter.

        Parameters
        ----------
        hyperparameter : FloatHyperparameter
            A float hyperparameter that, due to using a random forest surrogate, must have its prior discretized.
        X_col: np.ndarray [N, ]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, ),
            with N as the number of points to evaluate for the specific hyperparameter.
        number_of_bins : int
            The number of unique values allowed on the discretized version of the pdf.

        Returns
        -------
        np.ndarray [N, 1]
            The user prior over the optimum for the parameter at hand.
        """
        # Evaluates the actual pdf on all the relevant points
        # Replace deprecated method
        pdf_values = hyperparameter._pdf(X_col[:, np.newaxis])

        # Retrieves the largest value of the pdf in the domain
        lower, upper = (0, hyperparameter.get_max_density())

        # Creates the bins (the possible discrete options of the pdf)
        bin_values = np.linspace(lower, upper, number_of_bins)

        # Generates an index (bin) for each evaluated point
        bin_indices = np.clip(
            np.round((pdf_values - lower) * number_of_bins / (upper - lower)), 0, number_of_bins - 1
        ).astype(int)

        # Gets the actual value for each point
        prior_values = bin_values[bin_indices]

        return prior_values

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute the prior-weighted acquisition function values, where the prior on each
        parameter is multiplied by a decay factor controlled by the parameter decay_beta and
        the iteration number. Multivariate priors are not supported, for now.

        Parameters
        ----------
        X: np.ndarray [N, D]
            The input points where the acquisition function should be evaluated. The dimensionality of X is (N, D),
            with N as the number of points to evaluate at and D is the number of dimensions of one X.

        Returns
        -------
        np.ndarray [N, 1]
            Prior-weighted acquisition function values of X
        """
        if self._rescale:
            # for TS and UCB, we need to scale the function values to not run into issues
            # of negative values or issues of varying magnitudes (here, they are both)
            # negative by design and just flipping the sign leads to picking the worst point)
            acq_values = np.clip(self._acquisition_function._compute(X) + self._eta, 0, np.inf)
        else:
            acq_values = self._acquisition_function._compute(X)

        prior_values = self._compute_prior(X) + self._prior_floor
        decayed_prior_values = np.power(prior_values, self._decay_beta / (self._iteration_number + 1))

        return acq_values * decayed_prior_values



from __future__ import annotations

from typing import Any, Iterable, Iterator, Mapping, cast

import json
from collections import OrderedDict
from pathlib import Path

import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace

from smac.constants import MAXINT
from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.runhistory.dataclasses import (
    InstanceSeedBudgetKey,
    InstanceSeedKey,
    TrialInfo,
    TrialKey,
    TrialValue,
)
from smac.runhistory.enumerations import StatusType
from smac.utils.configspace import get_config_hash
from smac.utils.logging import get_logger
from smac.utils.multi_objective import normalize_costs
from smac.utils.numpyencoder import NumpyEncoder

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class RunHistory(Mapping[TrialKey, TrialValue]):
    """Container for the target function run information.

    Most importantly, the runhistory contains an efficient mapping from each evaluated configuration to the
    empirical cost observed on either the full instance set or a subset. The cost is the average over all
    observed costs for one configuration:

    * If using budgets for a single instance, only the cost on the highest observed budget is returned.
    * If using instances as the budget, the average cost over all evaluated instances is returned.
    * Theoretically, the runhistory object can handle instances and budgets at the same time. This is
      neither used nor tested.

    Note
    ----
    Guaranteed to be picklable.

    Parameters
    ----------
    multi_objective_algorithm : AbstractMultiObjectiveAlgorithm | None, defaults to None
        The multi-objective algorithm is required to scalarize the costs in case of multi-objective.
    overwrite_existing_trials : bool, defaults to false
        Overwrites a trial (combination of configuration, instance, budget and seed) if it already exists.
    """

    def __init__(
        self,
        multi_objective_algorithm: AbstractMultiObjectiveAlgorithm | None = None,
        overwrite_existing_trials: bool = False,
    ) -> None:
        self._multi_objective_algorithm = multi_objective_algorithm
        self._overwrite_existing_trials = overwrite_existing_trials
        self.reset()

    @property
    def submitted(self) -> int:
        """Returns how many trials have been submitted."""
        return self._submitted

    @property
    def finished(self) -> int:
        """Returns how many trials have been finished."""
        return self._finished

    @property
    def running(self) -> int:
        """Returns how many trials are still running."""
        return self._running

    @property
    def multi_objective_algorithm(self) -> AbstractMultiObjectiveAlgorithm | None:
        """The multi-objective algorithm required to scaralize the costs in case of multi-objective."""
        return self._multi_objective_algorithm

    @multi_objective_algorithm.setter
    def multi_objective_algorithm(self, value: AbstractMultiObjectiveAlgorithm) -> None:
        """We want to have the option to change the multi objective algorithm."""
        self._multi_objective_algorithm = value

    @property
    def ids_config(self) -> dict[int, Configuration]:
        """Mapping from config id to configuration."""
        return self._ids_config

    @property
    def config_ids(self) -> dict[Configuration, int]:
        """Mapping from configuration to config id."""
        return self._config_ids

    @property
    def objective_bounds(self) -> list[tuple[float, float]]:
        """Returns the lower and upper bound of each objective."""
        return self._objective_bounds

    def reset(self) -> None:
        """Resets this runhistory to its default state."""
        # By having the data in a deterministic order we can do useful tests when we
        # serialize the data and can assume it is still in the same order as it was added.
        self._data: dict[TrialKey, TrialValue] = OrderedDict()

        # Keep track of trials
        self._submitted = 0
        self._finished = 0
        self._running = 0

        # For fast access, we have also an unordered data structure to get all instance
        # seed pairs of a configuration.
        self._config_id_to_isk_to_budget: dict[int, dict[InstanceSeedKey, list[float | None]]] = {}
        self._running_trials: list[TrialInfo] = []

        self._config_ids: dict[Configuration, int] = {}
        self._ids_config: dict[int, Configuration] = {}
        self._n_id = 0

        # Stores cost for each configuration ID
        self._cost_per_config: dict[int, float | list[float]] = {}
        # Stores min cost across all budgets for each configuration ID
        self._min_cost_per_config: dict[int, float | list[float]] = {}
        # Maps the configuration ID to the number of runs for that configuration
        # and is necessary for computing the moving average.
        self._num_trials_per_config: dict[int, int] = {}

        # Store whether a datapoint is "external", which means it was read from
        # a JSON file. Can be chosen to not be written to disk.
        self._n_objectives: int = -1
        self._objective_bounds: list[tuple[float, float]] = []

    def __contains__(self, k: object) -> bool:
        """Dictionary semantics for `k in runhistory`."""
        return k in self._data

    def __getitem__(self, k: TrialKey) -> TrialValue:
        """Dictionary semantics for `v = runhistory[k]`."""
        return self._data[k]

    def __iter__(self) -> Iterator[TrialKey]:
        """Dictionary semantics for `for k in runhistory.keys()`."""
        return iter(self._data.keys())

    def __len__(self) -> int:
        """Enables the `len(runhistory)`"""
        return len(self._data)

    def __eq__(self, other: Any) -> bool:
        """Enables to check equality of runhistory if the run is continued."""
        return self._data == other._data

    def empty(self) -> bool:
        """Check whether the RunHistory is empty.

        Returns
        -------
        emptiness: bool
            True if trials have been added to the RunHistory.
        """
        return len(self._data) == 0

    def add(
        self,
        config: Configuration,
        cost: int | float | list[int | float],
        time: float = 0.0,
        cpu_time: float = 0.0,
        status: StatusType = StatusType.SUCCESS,
        instance: str | None = None,
        seed: int | None = None,
        budget: float | None = None,
        starttime: float = 0.0,
        endtime: float = 0.0,
        additional_info: dict[str, Any] = None,
        force_update: bool = False,
    ) -> None:
        """Adds a new trial to the RunHistory.

        Parameters
        ----------
        config : Configuration
        cost : int | float | list[int | float]
            Cost of the evaluated trial. Might be a list in case of multi-objective.
        time : float
            How much time was needed to evaluate the trial.
        cpu_time : float
            How much time was needed on the hardware to evaluate the trial.
        status : StatusType, defaults to StatusType.SUCCESS
            The status of the trial.
        instance : str | None, defaults to none
        seed : int | None, defaults to none
        budget : float | None, defaults to none
        starttime : float, defaults to 0.0
        endtime : float, defaults to 0.0
        additional_info : dict[str, Any], defaults to {}
        force_update : bool, defaults to false
            Overwrites a previous trial if the trial already exists.
        """
        if config is None:
            raise TypeError("Configuration must not be None.")
        elif not isinstance(config, Configuration):
            raise TypeError("Configuration is not of type Configuration, but %s." % type(config))
        if additional_info is None:
            additional_info = {}

        # Squeeze is important to reduce arrays with one element
        # to scalars.
        cost_array = np.asarray(cost).squeeze()
        n_objectives = np.size(cost_array)

        # Get the config id
        config_id = self._config_ids.get(config)

        if config_id is None:
            self._n_id += 1
            self._config_ids[config] = self._n_id
            self._ids_config[self._n_id] = config

            config_id = self._n_id

        # Set the id attribute of the config object, so that users can access it
        config.config_id = config_id

        if status != StatusType.RUNNING:
            if self._n_objectives == -1:
                self._n_objectives = n_objectives
            elif self._n_objectives != n_objectives:
                raise ValueError(
                    f"Cost is not of the same length ({n_objectives}) as the number of "
                    f"objectives ({self._n_objectives})."
                )

            # Let's always work with floats; Makes it easier to deal with later on
            # array.tolist(), it returns a scalar if the array has one element.
            c = cost_array.tolist()
            if self._n_objectives == 1:
                c = float(c)
            else:
                c = [float(i) for i in c]
        else:
            c = cost_array.tolist()

        if budget is not None:
            # Just to make sure we really add a float
            budget = float(budget)

        k = TrialKey(config_id=config_id, instance=instance, seed=seed, budget=budget)
        v = TrialValue(
            cost=c,
            time=time,
            cpu_time=cpu_time,
            status=status,
            starttime=starttime,
            endtime=endtime,
            additional_info=additional_info,
        )

        # Construct keys and values for the data dictionary
        for key, value in (
            ("config", dict(config)),
            ("config_id", config_id),
            ("instance", instance),
            ("seed", seed),
            ("budget", budget),
            ("cost", c),
            ("time", time),
            ("cpu_time", cpu_time),
            ("status", status),
            ("starttime", starttime),
            ("endtime", endtime),
            ("additional_info", additional_info),
            ("origin", config.origin),
        ):
            self._check_json_serializable(key, value, k, v)

        # Each trial_key is supposed to be used only once. Repeated tries to add
        # the same trial_key will be ignored silently if not capped.
        previous_k = self._data.get(k)
        if self._overwrite_existing_trials or force_update or previous_k is None:
            # Update stati
            if previous_k is None:
                if status == StatusType.RUNNING:
                    self._running += 1
                else:
                    self._finished += 1

                self._submitted += 1
            else:
                if previous_k.status == StatusType.RUNNING and status != StatusType.RUNNING:
                    self._running -= 1
                    self._finished += 1

            self._add(k, v, status)
        else:
            logger.info("Entry was not added to the runhistory because existing trials will not be overwritten.")

    def add_trial(self, info: TrialInfo, value: TrialValue) -> None:
        """Adds a trial to the runhistory.

        Parameters
        ----------
        trial : TrialInfo
            The ``TrialInfo`` object of the running trial.
        """
        self.add(
            config=info.config,
            cost=value.cost,
            time=value.time,
            cpu_time=value.cpu_time,
            status=value.status,
            instance=info.instance,
            seed=info.seed,
            budget=info.budget,
            starttime=value.starttime,
            endtime=value.endtime,
            additional_info=value.additional_info,
        )

    def add_running_trial(self, trial: TrialInfo) -> None:
        """Adds a running trial to the runhistory.

        Parameters
        ----------
        trial : TrialInfo
            The ``TrialInfo`` object of the running trial.
        """
        self.add(
            config=trial.config,
            cost=float(MAXINT),
            time=0.0,
            cpu_time=0.0,
            status=StatusType.RUNNING,
            instance=trial.instance,
            seed=trial.seed,
            budget=trial.budget,
        )

    def update_cost(self, config: Configuration) -> None:
        """Stores the performance of a configuration across the instances in `self._cost_per_config`
        and also updates `self._num_trials_per_config`.

        Parameters
        ----------
        config: Configuration
            configuration to update cost based on all trials in runhistory
        """
        config_id = self._config_ids[config]

        # Removing duplicates while keeping the order
        inst_seed_budgets = list(
            dict.fromkeys(self.get_instance_seed_budget_keys(config, highest_observed_budget_only=True))
        )
        self._cost_per_config[config_id] = self.average_cost(config, inst_seed_budgets)
        self._num_trials_per_config[config_id] = len(inst_seed_budgets)

        all_isb = list(dict.fromkeys(self.get_instance_seed_budget_keys(config, highest_observed_budget_only=False)))
        self._min_cost_per_config[config_id] = self.min_cost(config, all_isb)

    def incremental_update_cost(self, config: Configuration, cost: float | list[float]) -> None:
        """Incrementally updates the performance of a configuration by using a moving average.

        Parameters
        ----------
        config: Configuration
            configuration to update cost based on all trials in runhistory
        cost: float
            cost of new run of config
        """
        config_id = self._config_ids[config]
        n_trials = self._num_trials_per_config.get(config_id, 0)

        if self._n_objectives > 1:
            costs = np.array(cost)
            old_costs = self._cost_per_config.get(config_id, np.array([0.0 for _ in range(self._n_objectives)]))
            old_costs = np.array(old_costs)

            new_costs = ((old_costs * n_trials) + costs) / (n_trials + 1)
            self._cost_per_config[config_id] = new_costs.tolist()
        else:
            old_cost = self._cost_per_config.get(config_id, 0.0)

            assert isinstance(cost, float)
            assert isinstance(old_cost, float)
            self._cost_per_config[config_id] = ((old_cost * n_trials) + cost) / (n_trials + 1)

        self._num_trials_per_config[config_id] = n_trials + 1

    def get_cost(self, config: Configuration) -> float:
        """Returns empirical cost for a configuration. See the class docstring for how the costs are
        computed. The costs are not re-computed, but are read from cache.

        Parameters
        ----------
        config: Configuration

        Returns
        -------
        cost: float
            Computed cost for configuration
        """
        config_id = self._config_ids.get(config)

        # Cost is always a single value (Single objective) or a list of values (Multi-objective)
        # For example, _cost_per_config always holds the value on the highest budget
        cost = self._cost_per_config.get(config_id, np.nan)  # type: ignore[arg-type] # noqa F821

        if self._n_objectives > 1:
            assert isinstance(cost, list)
            assert self.multi_objective_algorithm is not None

            # We have to normalize the costs here
            costs = normalize_costs(cost, self._objective_bounds)

            # After normalization, we get the weighted average
            return self.multi_objective_algorithm(costs)

        assert isinstance(cost, float)
        return float(cost)

    def get_min_cost(self, config: Configuration) -> float:
        """Returns the lowest empirical cost for a configuration across all trials.

        See the class docstring for how the costs are computed. The costs are not re-computed
        but are read from cache.

        Parameters
        ----------
        config : Configuration

        Returns
        -------
        min_cost: float
            Computed cost for configuration
        """
        config_id = self._config_ids.get(config)
        cost = self._min_cost_per_config.get(config_id, np.nan)  # type: ignore

        if self._n_objectives > 1:
            assert isinstance(cost, list)
            assert self.multi_objective_algorithm is not None

            costs = normalize_costs(cost, self._objective_bounds)

            # Note: We have to mean here because we already got the min cost
            return self.multi_objective_algorithm(costs)

        assert isinstance(cost, float)
        return float(cost)

    def average_cost(
        self,
        config: Configuration,
        instance_seed_budget_keys: list[InstanceSeedBudgetKey] | None = None,
        normalize: bool = False,
    ) -> float | list[float]:
        """Return the average cost of a configuration. This is the mean of costs of all instance-
        seed pairs.

        Parameters
        ----------
        config : Configuration
            Configuration to calculate objective for.
        instance_seed_budget_keys : list, optional (default=None)
            List of tuples of instance-seeds-budget keys. If None, the runhistory is
            queried for all trials of the given configuration.
        normalize : bool, optional (default=False)
            Normalizes the costs wrt. objective bounds in the multi-objective setting.
            Only a float is returned if normalize is True. Warning: The value can change
            over time because the objective bounds are changing. Also, the objective weights are
            incorporated.

        Returns
        -------
        Cost: float | list[float]
            Average cost. In case of multiple objectives, the mean of each objective is returned.
        """
        costs = self._cost(config, instance_seed_budget_keys)
        if costs:
            if self._n_objectives > 1:
                # Each objective is averaged separately
                # [[100, 200], [0, 0]] -> [50, 100]
                averaged_costs = np.mean(costs, axis=0).tolist()

                if normalize:
                    assert self.multi_objective_algorithm is not None
                    normalized_costs = normalize_costs(averaged_costs, self._objective_bounds)

                    return self.multi_objective_algorithm(normalized_costs)
                else:
                    return averaged_costs

            return float(np.mean(costs))

        return np.nan

    def sum_cost(
        self,
        config: Configuration,
        instance_seed_budget_keys: list[InstanceSeedBudgetKey] | None = None,
        normalize: bool = False,
    ) -> float | list[float]:
        """Return the sum of costs of a configuration. This is the sum of costs of all instance-seed
        pairs.

        Parameters
        ----------
        config : Configuration
            Configuration to calculate objective for.
        instance_seed_budget_keys : list, optional (default=None)
            List of tuples of instance-seeds-budget keys. If None, the runhistory is
            queried for all trials of the given configuration.
        normalize : bool, optional (default=False)
            Normalizes the costs wrt objective bounds in the multi-objective setting.
            Only a float is returned if normalize is True. Warning: The value can change
            over time because the objective bounds are changing. Also, the objective weights are
            incorporated.

        Returns
        -------
        sum_cost: float | list[float]
            Sum of costs of config. In case of multiple objectives, the costs are summed up for each
            objective individually.
        """
        costs = self._cost(config, instance_seed_budget_keys)
        if costs:
            if self._n_objectives > 1:
                # Each objective is summed separately
                # [[100, 200], [20, 10]] -> [120, 210]
                summed_costs = np.sum(costs, axis=0).tolist()

                if normalize:
                    assert self.multi_objective_algorithm is not None
                    normalized_costs = normalize_costs(summed_costs, self._objective_bounds)

                    return self.multi_objective_algorithm(normalized_costs)
                else:
                    return summed_costs

        return float(np.sum(costs))

    def min_cost(
        self,
        config: Configuration,
        instance_seed_budget_keys: list[InstanceSeedBudgetKey] | None = None,
        normalize: bool = False,
    ) -> float | list[float]:
        """Return the minimum cost of a configuration. This is the minimum cost of all instance-seed
         pairs.

        Warning
        -------
        In the case of multi-fidelity, the minimum cost per objectives is returned.

        Parameters
        ----------
        config : Configuration
            Configuration to calculate objective for.
        instance_seed_budget_keys : list, optional (default=None)
            List of tuples of instance-seeds-budget keys. If None, the runhistory is
            queried for all trials of the given configuration.
        normalize : bool, optional (default=False)
            Normalizes the costs wrt objective bounds in the multi-objective setting.
            Only a float is returned if normalize is True. Warning: The value can change
            over time because the objective bounds are changing. Also, the objective weights are
            incorporated.

        Returns
        -------
        min_cost: float | list[float]
            Minimum cost of the config. In case of multi-objective, the minimum cost per objective
            is returned.
        """
        costs = self._cost(config, instance_seed_budget_keys)
        if costs:
            if self._n_objectives > 1:
                # Each objective is viewed separately
                # [[100, 200], [20, 500]] -> [20, 200]
                min_costs = np.min(costs, axis=0).tolist()

                if normalize:
                    assert self.multi_objective_algorithm is not None
                    normalized_costs = normalize_costs(min_costs, self._objective_bounds)

                    return self.multi_objective_algorithm(normalized_costs)
                else:
                    return min_costs

            return float(np.min(costs))

        return np.nan

    def get_config(self, config_id: int) -> Configuration:
        """Returns the configuration from the configuration id."""
        return self._ids_config[config_id]

    def get_config_id(self, config: Configuration) -> int:
        """Returns the configuration id from a configuration."""
        return self._config_ids[config]

    def has_config(self, config: Configuration) -> bool:
        """Check if the config is stored in the runhistory"""
        return config in self._config_ids

    def get_configs(self, sort_by: str | None = None) -> list[Configuration]:
        """Return all configurations in this RunHistory object.

        Parameters
        ----------
        sort_by : str | None, defaults to None
            Sort the configs by ``cost`` (lowest cost first) or ``num_trials`` (config with lowest number of trials
            first).

        Returns
        -------
        configurations : list
            All configurations in the runhistory.
        """
        configs = list(self._config_ids.keys())

        if sort_by == "cost":
            return sorted(configs, key=lambda config: self._cost_per_config[self._config_ids[config]])
        elif sort_by == "num_trials":
            return sorted(configs, key=lambda config: len(self.get_trials(config)))
        elif sort_by is None:
            return configs
        else:
            raise ValueError(f"Unknown sort_by value: {sort_by}.")

    def get_configs_per_budget(
        self,
        budget_subset: list[float | int | None] | None = None,
    ) -> list[Configuration]:
        """Return all configs in this runhistory that have been run on one of these budgets.

        Parameters
        ----------
        budget_subset: list[float | int | None] | None, defaults to None

        Returns
        -------
        configurations : list
            List of configurations that have been run on the budgets in ``budget_subset``.
        """
        if budget_subset is None:
            return self.get_configs()

        configs = []
        for key in self._data.keys():
            if key.budget in budget_subset:
                configs.append(self._ids_config[key.config_id])

        return configs

    def get_running_configs(self) -> list[Configuration]:
        """Returns all configurations which have at least one running trial.

        Returns
        -------
        list[Configuration]
            List of configurations, all of which have at least one running trial.
        """
        configs = []
        for trial in self._running_trials:
            if trial.config not in configs:
                configs.append(trial.config)

        return configs

    def get_trials(
        self,
        config: Configuration,
        highest_observed_budget_only: bool = True,
    ) -> list[TrialInfo]:
        """Returns all trials for a configuration.

        Warning
        -------
        Does not return running trials. Please use ``get_running_trials`` to receive running trials.

        Parameters
        ----------
        config : Configuration
        highest_observed_budget_only : bool
            Select only the highest observed budget run for this configuration.
            Meaning on multiple executions of the same instance-seed pair for a
            a given configuration, only the highest observed budget is returned.

        Returns
        -------
        trials : list[InstanceSeedBudgetKey]
            List of trials for the passed configuration.
        """
        config_id = self._config_ids.get(config)
        trials = {}
        if config_id in self._config_id_to_isk_to_budget:
            trials = self._config_id_to_isk_to_budget[config_id].copy()

        # Select only the max budget run if specified
        if highest_observed_budget_only:
            for k, v in trials.items():
                if None in v:
                    trials[k] = [None]
                else:
                    trials[k] = [max([v_ for v_ in v if v_ is not None])]

        return [TrialInfo(config, k.instance, k.seed, budget) for k, v in trials.items() for budget in v]

    def get_running_trials(self, config: Configuration | None = None) -> list[TrialInfo]:
        """Returns all running trials for the passed configuration.

        Parameters
        ----------
        config : Configuration | None, defaults to None
            Return only running trials from the passed configuration. If None, all configs are
            considered.

        Returns
        -------
        trials : list[TrialInfo]
            List of trials, all of which are still running.
        """
        # Always work on copies
        if config is None:
            return [trial for trial in self._running_trials]
        else:
            return [trial for trial in self._running_trials if trial.config == config]

    def get_instance_seed_budget_keys(
        self,
        config: Configuration,
        highest_observed_budget_only: bool = True,
    ) -> list[InstanceSeedBudgetKey]:
        """
        Uses ``get_trials`` to return a list of instance-seed-budget keys.

        Warning
        -------
        Does not return running instances.

        Parameters
        ----------
        config : Configuration
        highest_observed_budget_only : bool, defaults to True
            Select only the highest observed budget run for this configuration.

        Returns
        -------
        list[InstanceSeedBudgetKey]
        """
        trials = self.get_trials(config, highest_observed_budget_only)

        # Convert to instance-seed-budget key
        return [InstanceSeedBudgetKey(t.instance, t.seed, t.budget) for t in trials]

    def save(self, filename: str | Path = "runhistory.json") -> None:
        """Saves RunHistory to disk.

        Parameters
        ----------
        filename : str | Path, defaults to "runhistory.json"
        """
        data = list()
        for k, v in self._data.items():
            data.append(
                {
                    "config_id": int(k.config_id),
                    "instance": str(k.instance) if k.instance is not None else None,
                    "seed": int(k.seed) if k.seed is not None else None,
                    "budget": float(k.budget) if k.budget is not None else None,
                    "cost": v.cost,
                    "time": v.time,
                    "cpu_time": v.cpu_time,
                    "status": v.status,
                    "starttime": v.starttime,
                    "endtime": v.endtime,
                    "additional_info": v.additional_info,
                }
            )

        config_ids_to_serialize = set([entry["config_id"] for entry in data])
        configs = {}
        config_origins = {}
        for id_, config in self._ids_config.items():
            if id_ in config_ids_to_serialize:
                configs[id_] = dict(config)

            config_origins[id_] = config.origin

        if isinstance(filename, str):
            filename = Path(filename)

        assert str(filename).endswith(".json")
        filename.parent.mkdir(parents=True, exist_ok=True)

        with open(filename, "w") as fp:
            assert self._running == len(self._running_trials)
            json.dump(
                {
                    "stats": {"submitted": self._submitted, "finished": self._finished, "running": self._running},
                    "data": data,
                    "configs": configs,
                    "config_origins": config_origins,
                },
                fp,
                indent=2,
                cls=NumpyEncoder,
            )

    def load(self, filename: str | Path, configspace: ConfigurationSpace) -> None:
        """Loads the runhistory from disk.

        Warning
        -------
        Overwrites the current runhistory.

        Parameters
        ----------
        filename : str | Path
        configspace : ConfigSpace
        """
        if isinstance(filename, str):
            filename = Path(filename)

        # We reset the RunHistory first to avoid any inconsistencies
        self.reset()

        try:
            with open(filename) as fp:
                data = json.load(fp)
        except Exception as e:
            logger.warning(
                f"Encountered exception {e} while reading RunHistory from {filename}. Not adding any trials!"
            )
            return

        config_origins = data.get("config_origins", {})

        self._ids_config = {}
        for id_, values in data["configs"].items():
            self._ids_config[int(id_)] = Configuration(
                configspace,
                values=values,
                origin=config_origins.get(id_, None),
            )

        self._config_ids = {config: id_ for id_, config in self._ids_config.items()}
        self._n_id = len(self._config_ids)

        # Important to use add method to use all data structure correctly
        # NOTE: These hardcoded indices can easily lead to trouble
        for entry in data["data"]:
            if self._n_objectives == -1:
                if isinstance(entry["cost"], (float, int)):
                    self._n_objectives = 1
                else:
                    self._n_objectives = len(entry["cost"])

            cost: list[float] | float
            if self._n_objectives == 1:
                cost = float(entry["cost"])
            else:
                cost = [float(x) for x in entry["cost"]]
            self.add(
                config=self._ids_config[int(entry["config_id"])],
                cost=cost,
                time=entry["time"],
                cpu_time=entry["cpu_time"],
                status=StatusType(entry["status"]),
                instance=entry["instance"],
                seed=entry["seed"],
                budget=entry["budget"],
                starttime=entry["starttime"],
                endtime=entry["endtime"],
                additional_info=entry["additional_info"],
            )

        # Although adding trials should give us the same stats, the trajectory might be different
        # because of the running status and/or overwriting trials
        # Therefore, we just overwrite them
        self._submitted = data["stats"]["submitted"]
        self._finished = data["stats"]["finished"]
        self._running = data["stats"]["running"]

    def update_from_json(
        self,
        filename: str,
        configspace: ConfigurationSpace,
    ) -> None:
        """Updates the current RunHistory by adding new trials from a json file.

        Parameters
        ----------
        filename : str
            File name to load from.
        configspace : ConfigurationSpace
        """
        new_runhistory = RunHistory()
        new_runhistory.load(filename, configspace)
        self.update(runhistory=new_runhistory)

    def update(self, runhistory: RunHistory) -> None:
        """Updates the current RunHistory by adding new trials from another RunHistory.

        Parameters
        ----------
        runhistory : RunHistory
            RunHistory with additional data to be added to self
        """
        # Configurations might be already known, but by a different ID. This
        # does not matter here because the add() method handles this
        # correctly by assigning an ID to unknown configurations and re-using the ID.
        for key, value in runhistory.items():
            config = runhistory._ids_config[key.config_id]
            self.add(
                config=config,
                cost=value.cost,
                time=value.time,
                cpu_time=value.cpu_time,
                status=value.status,
                instance=key.instance,
                starttime=value.starttime,
                endtime=value.endtime,
                seed=key.seed,
                budget=key.budget,
                additional_info=value.additional_info,
            )

    def update_costs(self, instances: list[str] | None = None) -> None:
        """Computes the cost of all configurations from scratch and overwrites `self._cost_per_config`
        and `self._num_trials_per_config` accordingly.

        Parameters
        ----------
        instances: list[str] | None, defaults to none
            List of instances; if given, cost is only computed wrt to this instance set.
        """
        self._cost_per_config = {}
        self._num_trials_per_config = {}
        for config, config_id in self._config_ids.items():
            # Removing duplicates while keeping the order
            inst_seed_budgets = list(
                dict.fromkeys(self.get_instance_seed_budget_keys(config, highest_observed_budget_only=True))
            )
            if instances is not None:
                inst_seed_budgets = list(filter(lambda x: x.instance in cast(list, instances), inst_seed_budgets))

            if inst_seed_budgets:  # can be empty if never saw any trials on instances
                self._cost_per_config[config_id] = self.average_cost(config, inst_seed_budgets)
                self._min_cost_per_config[config_id] = self.min_cost(config, inst_seed_budgets)
                self._num_trials_per_config[config_id] = len(inst_seed_budgets)

    def _check_json_serializable(
        self,
        key: str,
        obj: Any,
        trial_key: TrialKey,
        trial_value: TrialValue,
    ) -> None:
        try:
            json.dumps(obj, cls=NumpyEncoder)
        except Exception as e:
            raise ValueError(
                "Cannot add %s: %s of type %s to runhistory because it raises an error during JSON encoding, "
                "please see the error above.\ntrial_key: %s\ntrial_value %s"
                % (key, str(obj), type(obj), trial_key, trial_value)
            ) from e

    def _update_objective_bounds(self) -> None:
        """Update the objective bounds based on the data in the RunHistory."""
        all_costs = []
        for run_value in self._data.values():
            costs = run_value.cost
            if run_value.status == StatusType.SUCCESS:
                if not isinstance(costs, Iterable):
                    costs = [costs]

                assert len(costs) == self._n_objectives
                all_costs.append(costs)

        all_costs = np.array(all_costs, dtype=float)  # type: ignore[assignment]

        if len(all_costs) == 0:
            self._objective_bounds = [(np.inf, -np.inf)] * self._n_objectives
            return

        min_values = np.min(all_costs, axis=0)
        max_values = np.max(all_costs, axis=0)

        self._objective_bounds = []
        for min_v, max_v in zip(min_values, max_values):
            self._objective_bounds += [(min_v, max_v)]

    def _add(self, k: TrialKey, v: TrialValue, status: StatusType) -> None:
        """
        Actual function to add new entry to data structures.

        Note
        ----
        This method always calls `update_cost` in the multi-objective setting.
        """
        self._data[k] = v

        # Update objective bounds based on raw data
        self._update_objective_bounds()

        # Do not register the cost until the run has completed
        if status != StatusType.RUNNING:
            # Also add to fast data structure
            isk = InstanceSeedKey(k.instance, k.seed)
            self._config_id_to_isk_to_budget[k.config_id] = self._config_id_to_isk_to_budget.get(k.config_id, {})

            # We sanity-check whether we don't mix none and str in the instances
            for isk_ in self._config_id_to_isk_to_budget[k.config_id].keys():
                if isinstance(isk_, str) != isinstance(isk, str):
                    raise ValueError(
                        "Can not mix instances of different types. "
                        f"Wants to add {isk_.instance} but found already {isk.instance}."
                    )

            if isk not in self._config_id_to_isk_to_budget[k.config_id]:
                # Add new inst-seed-key with budget to main dict
                self._config_id_to_isk_to_budget[k.config_id][isk] = [k.budget]
            # Before it was k.budget not in isk
            elif k.budget != isk.instance and k.budget != isk.seed:
                # We have to make sure that we don't mix none and float budgets
                if isinstance(self._config_id_to_isk_to_budget[k.config_id][isk][0], float) != isinstance(
                    k.budget, float
                ):
                    raise ValueError(
                        "Can not mix budgets of different types for the same instance-seed pair. "
                        f"Wants to add {k.budget} but found already "
                        f"{self._config_id_to_isk_to_budget[k.config_id][isk][0]}."
                    )

                # Append new budget to existing inst-seed-key dict
                self._config_id_to_isk_to_budget[k.config_id][isk].append(k.budget)

            config = self._ids_config[k.config_id]
            config_hash = get_config_hash(config)

            # If budget is used, then update cost instead of incremental updates
            if not self._overwrite_existing_trials and k.budget == 0:
                logger.debug(f"Incremental update cost for config {config_hash}")
                # Assumes an average across trials as cost function aggregation, this is used for
                # algorithm configuration (incremental updates are used to save time as getting the
                # cost for > 100 instances is high)
                self.incremental_update_cost(config, v.cost)
            else:
                # This happens when budget > 0 (only successive halving and hyperband so far)
                logger.debug(f"Update cost for config {config_hash}.")
                self.update_cost(config)

        # Make TrialInfo object
        trial_info = TrialInfo(self.get_config(k.config_id), instance=k.instance, seed=k.seed, budget=k.budget)

        # Fast data structure for pending trials
        if status == StatusType.RUNNING:
            # Add to running cache
            self._running_trials.append(trial_info)
        else:
            # Remove from cache
            if trial_info in self._running_trials:
                self._running_trials.remove(trial_info)

    def _cost(
        self,
        config: Configuration,
        instance_seed_budget_keys: list[InstanceSeedBudgetKey] | None = None,
    ) -> list[float | list[float]]:
        """Returns a list of all costs for the given config for further calculations.
        The costs are directly taken from the RunHistory data.

        Parameters
        ----------
        config : Configuration
            Configuration to calculate objective for.
        instance_seed_budget_keys : list, defaults to None
            List of tuples of instance-seeds-budget keys. If None, the RunHistory is
            queried for all trials of the given configuration.

        Returns
        -------
        costs: list[list[float] | list[list[float]]]
            List of all found costs. In case of multi-objective, the list contains lists.
        """
        try:
            id_ = self._config_ids[config]
        except KeyError:  # Challenger was not running so far
            return []

        if instance_seed_budget_keys is None:
            instance_seed_budget_keys = self.get_instance_seed_budget_keys(config, highest_observed_budget_only=True)

        costs = []
        for key in instance_seed_budget_keys:
            k = TrialKey(
                config_id=id_,
                instance=key.instance,
                seed=key.seed,
                budget=key.budget,
            )

            costs.append(self._data[k].cost)

        return costs



from smac.facade.abstract_facade import AbstractFacade
from smac.facade.algorithm_configuration_facade import AlgorithmConfigurationFacade
from smac.facade.blackbox_facade import BlackBoxFacade
from smac.facade.hyperband_facade import HyperbandFacade
from smac.facade.hyperparameter_optimization_facade import (
    HyperparameterOptimizationFacade,
)
from smac.facade.multi_fidelity_facade import MultiFidelityFacade
from smac.facade.random_facade import RandomFacade

__all__ = [
    "AbstractFacade",
    "AlgorithmConfigurationFacade",
    "BlackBoxFacade",
    "HyperparameterOptimizationFacade",
    "MultiFidelityFacade",
    "HyperbandFacade",
    "RandomFacade",
]



from __future__ import annotations

import json
import platform
from datetime import datetime

import smac
from smac.callback.callback import Callback
from smac.main.smbo import SMBO
from smac.utils.numpyencoder import NumpyEncoder

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class MetadataCallback(Callback):
    def __init__(self, **kwargs: str | int | float | dict | list) -> None:
        # Arguments must be json serializable
        self.kwargs = kwargs

    def on_start(self, smbo: SMBO) -> None:
        """Called before the optimization starts."""
        path = smbo._scenario.output_directory
        meta_dict = {
            "utc_time": str(datetime.utcnow()),
            "os": platform.platform(),
            "smac_version": getattr(smac, "version"),
        }
        for key, value in self.kwargs.items():
            meta_dict[key] = value

        path.mkdir(parents=True, exist_ok=True)

        with open(path / "metadata.json", "w") as fp:
            json.dump(meta_dict, fp, indent=2, cls=NumpyEncoder)



from smac.runhistory.encoder.abstract_encoder import AbstractRunHistoryEncoder

# from smac.runhistory.encoder.boing_encoder import (
#     RunHistoryRawEncoder,
#     RunHistoryRawScaledEncoder,
# )
from smac.runhistory.encoder.eips_encoder import RunHistoryEIPSEncoder
from smac.runhistory.encoder.encoder import RunHistoryEncoder
from smac.runhistory.encoder.inverse_scaled_encoder import (
    RunHistoryInverseScaledEncoder,
)
from smac.runhistory.encoder.log_encoder import RunHistoryLogEncoder
from smac.runhistory.encoder.log_scaled_encoder import RunHistoryLogScaledEncoder
from smac.runhistory.encoder.scaled_encoder import RunHistoryScaledEncoder
from smac.runhistory.encoder.sqrt_scaled_encoder import RunHistorySqrtScaledEncoder

__all__ = [
    "AbstractRunHistoryEncoder",
    "RunHistoryEncoder",
    # "RunHistoryRawEncoder",
    # "RunHistoryRawScaledEncoder",
    "RunHistoryEIPSEncoder",
    "RunHistoryInverseScaledEncoder",
    "RunHistoryLogEncoder",
    "RunHistoryLogScaledEncoder",
    "RunHistoryScaledEncoder",
    "RunHistorySqrtScaledEncoder",
]



from pathlib import Path


def replace(file: str, find: str, replace: str) -> None:
    new_lines = []
    with open(file) as f:
        for line in f.readlines():
            line = line.replace(find, replace)
            new_lines += [line]

    # Empty file
    with open(file, "w") as f:
        f.write("")

    # Save file with new lines
    with open(file, "a") as f:
        for line in new_lines:
            f.write(line)


def replace_if_starts_with(file: str, starts_with: str, replace: str = "") -> None:
    new_lines = []
    with open(file) as f:
        for line in f.readlines():
            if line.startswith(starts_with):
                line = replace
                if len(replace) > 0:
                    line = line + "\n"

            new_lines += [line]

    # Empty file
    with open(file, "w") as f:
        f.write("")

    # Save file with new lines
    with open(file, "a") as f:
        for line in new_lines:
            f.write(line)


def check(file: str) -> None:
    has_copyright = False
    has_license = False

    copyright = "__copyright__"
    license = "__license__"

    with open(file) as f:
        for line in f.readlines():
            if line.startswith(copyright):
                has_copyright = True

            if line.startswith(license):
                has_license = True

    if not has_copyright:
        print(f"{file} does not have copyright.")

    if not has_license:
        print(f"{file} does not have license.")


if __name__ == "__main__":

    for path in Path("smac").rglob("*.py"):
        filename = str(path)
        if filename == "smac/__init__.py":
            continue

        replace_if_starts_with(filename, "__copyright__ =", '__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"')
        replace_if_starts_with(filename, "__author__ =")
        replace_if_starts_with(filename, "__maintainer__ =")
        replace_if_starts_with(filename, "__version__ =")
        replace_if_starts_with(filename, "__email__ =")
        check(filename)



from __future__ import annotations

from typing import Any

import numpy as np
import scipy.stats as sps

from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class GammaPrior(AbstractPrior):
    """Implementation of gamma prior.

    f(x) = (x-loc)**(a-1) * e**(-(x-loc)) * (1/scale)**a / gamma(a)

    Parameters
    ----------
    a : float
        The shape parameter. Must be greater than 0.
    scale : float
        The scale parameter (1/scale corresponds to parameter p in canonical form). Must be greather than 0.
    loc : float
        Mean parameter for the distribution.
    seed : int, defaults to 0
    """

    def __init__(
        self,
        a: float,
        scale: float,
        loc: float,
        seed: int = 0,
    ):
        super().__init__(seed=seed)
        assert a > 0
        assert scale > 0

        self._a = a
        self._loc = loc
        self._scale = scale

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "a": self._a,
                "loc": self._loc,
                "scale": self._scale,
            }
        )

        return meta

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        return self._rng.gamma(shape=self._a, scale=self._scale, size=n_samples)

    def _get_log_probability(self, theta: float) -> float:
        """Returns the log pdf of theta."""
        if np.ndim(theta) != 0:
            raise NotImplementedError()

        return sps.gamma.logpdf(theta, a=self._a, scale=self._scale, loc=self._loc)

    def _get_gradient(self, theta: float) -> float:
        """Get gradient as computed by Wolfram Alpha."""
        if np.ndim(theta) == 0:
            # Multiply by theta because of the chain rule...
            return ((self._a - 1) / theta - (1 / self._scale)) * theta
        else:
            raise NotImplementedError()



from __future__ import annotations

import pandas as pd
from pathlib import Path
from src.datasets.dataset import Dataset


class ACBraninDataset(Dataset):
    def __init__(self) -> None:
        self._data = pd.read_csv(Path(__file__).parent / "ac_branin_instances_features.csv", skipinitialspace=True)
        self.data["instance"] = self.data["instance"].apply(str)

    def get_instances(self, n: int | None = None) -> list[str]:
        """Create instances from the dataset which include two classes only."""
        instances = self.data["instance"].to_list()

        if n is not None:
            instances = instances[:n]

        return instances

    def get_instance_features(self, n: int | None = None) -> dict[str, list[int | float]]:
        """Returns the mean and variance of all instances as features."""
        features = {}
        for instance in self.get_instances(n):
            features[instance] = [float(self.data["feature"][self.data["instance"] == instance])]

        return features



import datetime
import traceback

name = "SMAC3"
package_name = "smac"
author = (
    "\tMarius Lindauer, Katharina Eggensperger, Matthias Feurer, André Biedenkapp, "
    "Difan Deng,\n\tCarolin Benjamins, Tim Ruhkopf, René Sass and Frank Hutter"
)

author_email = "fh@cs.uni-freiburg.de"
description = "SMAC3, a Python implementation of 'Sequential Model-based Algorithm Configuration'."
url = "https://www.automl.org/"
project_urls = {
    "Documentation": "https://automl.github.io/SMAC3/main",
    "Source Code": "https://github.com/automl/SMAC3",
}
copyright = (
    f"Copyright {datetime.date.today().strftime('%Y')}, Marius Lindauer, Katharina Eggensperger, "
    "Matthias Feurer, André Biedenkapp, Difan Deng, Carolin Benjamins, Tim Ruhkopf, René Sass "
    "and Frank Hutter"
)
version = "2.3.0"


try:
    from smac.callback.callback import Callback
    from smac.facade import (
        AlgorithmConfigurationFacade,
        BlackBoxFacade,
        HyperbandFacade,
        HyperparameterOptimizationFacade,
        MultiFidelityFacade,
        RandomFacade,
    )
    from smac.runhistory.runhistory import RunHistory
    from smac.scenario import Scenario

    __all__ = [
        "Scenario",
        "RunHistory",
        "BlackBoxFacade",
        "HyperparameterOptimizationFacade",
        "MultiFidelityFacade",
        "AlgorithmConfigurationFacade",
        "RandomFacade",
        "HyperbandFacade",
        "Callback",
    ]
except ModuleNotFoundError as e:
    print(e)
    traceback.print_exc()



from smac.intensifier.abstract_intensifier import AbstractIntensifier
from smac.intensifier.hyperband import Hyperband
from smac.intensifier.intensifier import Intensifier
from smac.intensifier.successive_halving import SuccessiveHalving

__all__ = [
    "AbstractIntensifier",
    "Intensifier",
    "SuccessiveHalving",
    "Hyperband",
]



from __future__ import annotations

from typing import Any, Iterator

import time
from pathlib import Path

import dask
from ConfigSpace import Configuration
from dask.distributed import Client, Future, wait

from smac.runhistory import StatusType, TrialInfo, TrialValue
from smac.runner.abstract_runner import AbstractRunner
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class DaskParallelRunner(AbstractRunner):
    """Interface to submit and collect a job in a distributed fashion. DaskParallelRunner is
    intended to comply with the bridge design pattern. Nevertheless, to reduce the amount of code
    within single-vs-parallel implementations, DaskParallelRunner wraps a BaseRunner object which
    is then executed in parallel on `n_workers`.

    This class then is constructed by passing an AbstractRunner that implements
    a `run` method, and is capable of doing so in a serial fashion. Next,
    this wrapper class uses dask to initialize `N` number of AbstractRunner that actively wait of a
    TrialInfo to produce a RunInfo object.

    To be more precise, the work model is then:

    1. The intensifier dictates "what" to run (a configuration/instance/seed) via a TrialInfo object.
    2. An abstract runner takes this TrialInfo object and launches the task via
       `submit_run`. In the case of DaskParallelRunner, `n_workers` receive a pickle-object of
       `DaskParallelRunner.single_worker`, each with a `run` method coming from
       `DaskParallelRunner.single_worker.run()`
    3. TrialInfo objects are run in a distributed fashion, and their results are available locally to each worker. The
       result is collected by `iter_results` and then passed to SMBO.
    4. Exceptions are also locally available to each worker and need to be collected.

    Dask works with `Future` object which are managed via the DaskParallelRunner.client.

    Parameters
    ----------
    single_worker : AbstractRunner
        A runner to run in a distributed fashion. Will be distributed using `n_workers`.
    patience: int, default to 5
        How much to wait for workers (seconds) to be available if one fails.
    dask_client: Client | None, defaults to None
        User-created dask client, which can be used to start a dask cluster and then attach SMAC to it. This will not
        be closed automatically and will have to be closed manually if provided explicitly. If none is provided
        (default), a local one will be created for you and closed upon completion.
    """

    def __init__(
        self,
        single_worker: AbstractRunner,
        patience: int = 5,
        dask_client: Client | None = None,
    ):
        super().__init__(
            scenario=single_worker._scenario,
            required_arguments=single_worker._required_arguments,
        )

        # The single worker to hold on to and call run on
        self._single_worker = single_worker

        # The list of futures that dask will use to indicate in progress runs
        self._pending_trials: list[Future] = []

        # Dask related variables
        self._scheduler_file: Path | None = None
        self._patience = patience

        self._client: Client
        self._close_client_at_del: bool

        if dask_client is None:
            dask.config.set({"distributed.worker.daemon": False})
            self._close_client_at_del = True
            self._client = Client(
                n_workers=self._scenario.n_workers,
                processes=True,
                threads_per_worker=1,
                local_directory=str(self._scenario.output_directory),
            )

            if self._scenario.output_directory is not None:
                self._scheduler_file = Path(self._scenario.output_directory, ".dask_scheduler_file")
                self._client.write_scheduler_file(scheduler_file=str(self._scheduler_file))
        else:
            # We just use their set up
            self._client = dask_client
            self._close_client_at_del = False

    def submit_trial(self, trial_info: TrialInfo, **dask_data_to_scatter: dict[str, Any]) -> None:
        """This function submits a configuration embedded in a ``trial_info`` object, and uses one of
        the workers to produce a result locally to each worker.

        The execution of a configuration follows this procedure:

        #. The SMBO/intensifier generates a `TrialInfo`.
        #. SMBO calls `submit_trial` so that a worker launches the `trial_info`.
        #. `submit_trial` internally calls ``self.run()``. It does so via a call to `run_wrapper` which contains common
           code that any `run` method will otherwise have to implement.

        All results will be only available locally to each worker, so the main node needs to collect them.

        Parameters
        ----------
        trial_info : TrialInfo
            An object containing the configuration launched.

        dask_data_to_scatter: dict[str, Any]
            When a user scatters data from their local process to the distributed network,
            this data is distributed in a round-robin fashion grouping by number of cores.
            Roughly speaking, we can keep this data in memory and then we do not have to (de-)serialize the data
            every time we would like to execute a target function with a big dataset.
            For example, when your target function has a big dataset shared across all the target function,
            this argument is very useful.
        """
        # Check for resources or block till one is available
        if self.count_available_workers() <= 0:
            logger.debug("No worker available. Waiting for one to be available...")
            wait(self._pending_trials, return_when="FIRST_COMPLETED")
            self._process_pending_trials()

        # Check again to make sure that there are resources
        if self.count_available_workers() <= 0:
            logger.warning("No workers are available. This could mean workers crashed. Waiting for new workers...")
            time.sleep(self._patience)
            if self.count_available_workers() <= 0:
                raise RuntimeError(
                    "Tried to execute a job, but no worker was ever available."
                    "This likely means that a worker crashed or no workers were properly configured."
                )

        # At this point we can submit the job
        trial = self._client.submit(self._single_worker.run_wrapper, trial_info=trial_info, **dask_data_to_scatter)
        self._pending_trials.append(trial)

    def iter_results(self) -> Iterator[tuple[TrialInfo, TrialValue]]:  # noqa: D102
        self._process_pending_trials()
        while self._results_queue:
            yield self._results_queue.pop(0)

    def wait(self) -> None:  # noqa: D102
        if self.is_running():
            wait(self._pending_trials, return_when="FIRST_COMPLETED")

    def is_running(self) -> bool:  # noqa: D102
        return len(self._pending_trials) > 0

    def run(
        self,
        config: Configuration,
        instance: str | None = None,
        budget: float | None = None,
        seed: int | None = None,
        **dask_data_to_scatter: dict[str, Any],
    ) -> tuple[StatusType, float | list[float], float, float, dict]:  # noqa: D102
        return self._single_worker.run(
            config=config, instance=instance, seed=seed, budget=budget, **dask_data_to_scatter
        )

    def count_available_workers(self) -> int:
        """Total number of workers available. This number is dynamic as more resources
        can be allocated.
        """
        return sum(self._client.nthreads().values()) - len(self._pending_trials)

    def close(self, force: bool = False) -> None:
        """Closes the client."""
        if self._close_client_at_del or force:
            self._client.close()

    def _process_pending_trials(self) -> None:
        """The completed trials are moved from ``self._pending_trials`` to ``self._results_queue``.
        We make sure pending trials never exceed the capacity of the scheduler.
        """
        # In code check to make sure we don't exceed resource allocation
        if self.count_available_workers() < 0:
            logger.warning(
                "More running jobs than resources available. "
                "Should not have more pending trials in remote workers "
                "than the number of workers. This could mean a worker "
                "crashed and was not able to be recovered by dask. "
            )

        # Move the done run from the worker to the results queue
        done = [trial for trial in self._pending_trials if trial.done()]
        for trial in done:
            self._results_queue.append(trial.result())
            self._pending_trials.remove(trial)

    def __del__(self) -> None:
        """Makes sure that when this object gets deleted, the client is terminated. This
        is only done if the client was created by the dask runner.
        """
        if self._close_client_at_del:
            self.close()



from __future__ import annotations

from typing import Any, Callable

import copy
import inspect
import math
import time
import traceback
from functools import partial

import numpy as np
from ConfigSpace import Configuration
from pynisher import MemoryLimitException, WallTimeoutException, limit

from smac.runner.abstract_runner import StatusType
from smac.runner.abstract_serial_runner import AbstractSerialRunner
from smac.scenario import Scenario
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class TargetFunctionRunner(AbstractSerialRunner):
    """Class to execute target functions which are python functions. Evaluates function for given
    configuration and resource limit.

    The target function can either return a float (the loss), or a tuple with the first element
    being a float and the second being additional run information. In a multi-objective
    setting, the float value is replaced by a list of floats.

    Parameters
    ----------
    target_function : Callable
        The target function.
    scenario : Scenario
    required_arguments : list[str], defaults to []
        A list of required arguments, which are passed to the target function.
    """

    def __init__(
        self,
        scenario: Scenario,
        target_function: Callable,
        required_arguments: list[str] = None,
    ):
        if required_arguments is None:
            required_arguments = []
        super().__init__(scenario=scenario, required_arguments=required_arguments)
        self._target_function = target_function

        # Check if target function is callable
        if not callable(self._target_function):
            raise TypeError(
                "Argument `target_function` must be a callable but is type" f"`{type(self._target_function)}`."
            )

        # Signatures here
        signature = inspect.signature(self._target_function).parameters
        for argument in required_arguments:
            if argument not in signature.keys():
                raise RuntimeError(
                    f"Target function needs to have the arguments {required_arguments} "
                    f"but could not find {argument}."
                )

        # Now we check for additional arguments which are not used by SMAC
        # However, we only want to warn the user and not
        for key in list(signature.keys())[1:]:
            if key not in required_arguments:
                logger.warning(f"The argument {key} is not set by SMAC: Consider removing it from the target function.")

        # Pynisher limitations
        if (memory := self._scenario.trial_memory_limit) is not None:
            unit = None
            if isinstance(memory, (tuple, list)):
                memory, unit = memory
            memory = int(math.ceil(memory))
            if unit is not None:
                memory = (memory, unit)

        if (time := self._scenario.trial_walltime_limit) is not None:
            time = int(math.ceil(time))

        self._memory_limit = memory
        self._algorithm_walltime_limit = time

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta

        # Partial's don't have a __code__ attribute but are a convenient
        # way a user might want to pass a function to SMAC, specifying
        # keyword arguments.
        f = self._target_function
        if isinstance(f, partial):
            f = f.func
            meta.update({"code": str(f.__code__.co_code)})
            meta.update({"code-partial-args": repr(f)})
        else:
            meta.update({"code": str(self._target_function.__code__.co_code)})

        return meta

    def run(
        self,
        config: Configuration,
        instance: str | None = None,
        budget: float | None = None,
        seed: int | None = None,
        **dask_data_to_scatter: dict[str, Any],
    ) -> tuple[StatusType, float | list[float], float, float, dict]:
        """Calls the target function with pynisher if algorithm wall time limit or memory limit is
        set. Otherwise, the function is called directly.

        Parameters
        ----------
        config : Configuration
            Configuration to be passed to the target function.
        instance : str | None, defaults to None
            The Problem instance.
        budget : float | None, defaults to None
            A positive, real-valued number representing an arbitrary limit to the target function
            handled by the target function internally.
        seed : int, defaults to None
        dask_data_to_scatter: dict[str, Any]
            This kwargs must be empty when we do not use dask! ()
            When a user scatters data from their local process to the distributed network,
            this data is distributed in a round-robin fashion grouping by number of cores.
            Roughly speaking, we can keep this data in memory and then we do not have to (de-)serialize the data
            every time we would like to execute a target function with a big dataset.
            For example, when your target function has a big dataset shared across all the target function,
            this argument is very useful.

        Returns
        -------
        status : StatusType
            Status of the trial.
        cost : float | list[float]
            Resulting cost(s) of the trial.
        runtime : float
            The time the target function took to run.
        cpu_time : float
            The time the target function took on the hardware to run.
        additional_info : dict
            All further additional trial information.
        """
        # The kwargs are passed to the target function.
        kwargs: dict[str, Any] = {}
        kwargs.update(dask_data_to_scatter)

        if "seed" in self._required_arguments:
            kwargs["seed"] = seed

        if "instance" in self._required_arguments:
            kwargs["instance"] = instance

        if "budget" in self._required_arguments:
            kwargs["budget"] = budget

        # Presetting
        cost: float | list[float] = self._crash_cost
        runtime = 0.0
        cpu_time = runtime
        additional_info = {}
        status = StatusType.CRASHED

        # If memory limit or walltime limit is set, we wanna use pynisher
        target_function: Callable
        if self._memory_limit is not None or self._algorithm_walltime_limit is not None:
            target_function = limit(
                self._target_function,
                memory=self._memory_limit,
                wall_time=self._algorithm_walltime_limit,
                wrap_errors=True,  # Hard to describe; see https://github.com/automl/pynisher
            )
        else:
            target_function = self._target_function

        # We don't want the user to change the configuration
        config_copy = copy.deepcopy(config)

        # Call target function
        try:
            start_time = time.time()
            cpu_time = time.process_time()
            rval = self(config_copy, target_function, kwargs)
            cpu_time = time.process_time() - cpu_time
            runtime = time.time() - start_time
            status = StatusType.SUCCESS
        except WallTimeoutException:
            status = StatusType.TIMEOUT
        except MemoryLimitException:
            status = StatusType.MEMORYOUT
        except Exception as e:
            cost = np.asarray(cost).squeeze().tolist()
            additional_info = {
                "traceback": traceback.format_exc(),
                "error": repr(e),
            }
            status = StatusType.CRASHED

        if status != StatusType.SUCCESS:
            return status, cost, runtime, cpu_time, additional_info

        if isinstance(rval, tuple):
            result, additional_info = rval
        else:
            result, additional_info = rval, {}

        # Do some sanity checking (for multi objective)
        error = f"Returned costs {result} does not match the number of objectives {self._objectives}."

        # If dict convert to array and make sure the order is correct
        if isinstance(result, dict):
            if len(result) != len(self._objectives):
                raise RuntimeError(error)

            ordered_cost: list[float] = []
            for name in self._objectives:
                if name not in result:
                    raise RuntimeError(f"Objective {name} was not found in the returned costs.")  # noqa: E713

                ordered_cost.append(result[name])

            result = ordered_cost

        if isinstance(result, list):
            if len(result) != len(self._objectives):
                raise RuntimeError(error)

        if isinstance(result, float):
            if isinstance(self._objectives, list) and len(self._objectives) != 1:
                raise RuntimeError(error)

        cost = result

        if cost is None:
            status = StatusType.CRASHED
            cost = self._crash_cost

        # We want to get either a float or a list of floats.
        cost = np.asarray(cost).squeeze().tolist()

        return status, cost, runtime, cpu_time, additional_info

    def __call__(
        self,
        config: Configuration,
        algorithm: Callable,
        algorithm_kwargs: dict[str, Any],
    ) -> (
        float
        | list[float]
        | dict[str, float]
        | tuple[float, dict]
        | tuple[list[float], dict]
        | tuple[dict[str, float], dict]
    ):
        """Calls the algorithm, which is processed in the ``run`` method."""
        return algorithm(config, **algorithm_kwargs)



from __future__ import annotations

from typing import Any

import math

import numpy as np

from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class LogNormalPrior(AbstractPrior):
    """Implements the log normal prior.

    Parameters
    ----------
    sigma : float
        Specifies the standard deviation of the normal distribution.
    mean : float
        Specifies the mean of the normal distribution.
    seed : int, defaults to 0
    """

    def __init__(
        self,
        sigma: float,
        mean: float = 0,
        seed: int = 0,
    ):
        super().__init__(seed=seed)

        if mean != 0:
            raise NotImplementedError(mean)

        self._sigma = sigma
        self._sigma_square = sigma**2
        self._mean = mean
        self._sqrt_2_pi = np.sqrt(2 * np.pi)

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"sigma": self._sigma, "mean": self._mean})

        return meta

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        return self._rng.lognormal(mean=self._mean, sigma=self._sigma, size=n_samples)

    def _get_log_probability(self, theta: float) -> float:
        if theta <= self._mean:
            return -1e25
        else:
            rval = -((math.log(theta) - self._mean) ** 2) / (2 * self._sigma_square) - math.log(
                self._sqrt_2_pi * self._sigma * theta
            )
            return rval

    def _get_gradient(self, theta: float) -> float:
        if theta <= 0:
            return 0
        else:
            # Derivative of log(1 / (x * s^2 * sqrt(2 pi)) * exp( - 0.5 * (log(x ) / s^2))^2))
            # This is without the mean!
            return -(self._sigma_square + math.log(theta)) / (self._sigma_square * (theta)) * theta



from __future__ import annotations

from typing import Any

import numpy as np

from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.scenario import Scenario


class ParEGO(AbstractMultiObjectiveAlgorithm):
    """ParEGO implementation based on https://www.cs.bham.ac.uk/~jdk/UKCI-2015.pdf.

    Parameters
    ----------
    scenario : Scenario
    rho : float, defaults to 0.05
        A small positive value.
    seed : int | None, defaults to None
    """

    def __init__(
        self,
        scenario: Scenario,
        rho: float = 0.05,
        seed: int | None = None,
    ):
        super(ParEGO, self).__init__()

        if seed is None:
            seed = scenario.seed

        self._n_objectives = scenario.count_objectives()
        self._seed = seed
        self._rng = np.random.RandomState(seed)

        self._rho = rho
        # Will be set on starting an SMBO iteration
        self._theta: np.ndarray | None = None

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "name": self.__class__.__name__,
                "rho": self._rho,
                "seed": self._seed,
            }
        )

        return meta

    def update_on_iteration_start(self) -> None:  # noqa: D102
        self._theta = self._rng.rand(self._n_objectives)

        # Normalize so that all theta values sum up to 1
        self._theta = self._theta / (np.sum(self._theta) + 1e-10)

    def __call__(self, values: list[float]) -> float:  # noqa: D102
        # Weight the values
        if self._theta is None:
            raise ValueError("Iteration not yet initalized; Call `update_on_iteration_start()` first")

        theta_f = self._theta * values
        return float(np.max(theta_f, axis=0) + self._rho * np.sum(theta_f, axis=0))



from __future__ import annotations

from typing import Any, Iterator

from ConfigSpace import Configuration

from smac.intensifier.abstract_intensifier import AbstractIntensifier
from smac.runhistory import TrialInfo
from smac.runhistory.dataclasses import InstanceSeedBudgetKey
from smac.scenario import Scenario
from smac.utils.configspace import get_config_hash
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class Intensifier(AbstractIntensifier):
    """Implementation of an intensifier supporting multi-fidelity, multi-objective, and multi-processing.
    Races challengers against current incumbents.

    The behaviour of this intensifier is as follows:

    - First, adds configs from the runhistory to the queue with N=1 (they will be ignored if they are already
      evaluated).
    - While loop:

      - If queue is empty: Intensifies exactly one more instance of one incumbent and samples a new configuration
        afterwards.
      - If queue is not empty: Configs in the queue are evaluated on N=(N*2) instances if they might be better
        than the incumbents. If not, they are removed from the queue and rejected forever.

    Parameters
    ----------
    max_config_calls : int, defaults to 3
        Maximum number of configuration evaluations. Basically, how many instance-seed keys should be maxed evaluated
        for a configuration.
    max_incumbents : int, defaults to 10
        How many incumbents to keep track of in the case of multi-objective.
    retries : int, defaults to 16
        How many more iterations should be done in case no new trial is found.
    seed : int, defaults to None
        Internal seed used for random events, like shuffle seeds.
    """

    def __init__(
        self,
        scenario: Scenario,
        max_config_calls: int = 3,
        max_incumbents: int = 10,
        retries: int = 16,
        seed: int | None = None,
    ):
        super().__init__(scenario=scenario, max_config_calls=max_config_calls, max_incumbents=max_incumbents, seed=seed)
        self._retries = retries

    def reset(self) -> None:
        """Resets the internal variables of the intensifier including the queue."""
        super().reset()

        # Queue to keep track of the challengers
        # (config, N=how many trials should be sampled)
        self._queue: list[tuple[Configuration, int]] = []

    @property
    def uses_seeds(self) -> bool:  # noqa: D102
        return True

    @property
    def uses_budgets(self) -> bool:  # noqa: D102
        return False

    @property
    def uses_instances(self) -> bool:  # noqa: D102
        if self._scenario.instances is None:
            return False

        return True

    def get_state(self) -> dict[str, Any]:  # noqa: D102
        return {
            "queue": [
                (self.runhistory.get_config_id(config), n)
                for config, n in self._queue
                if self.runhistory.has_config(config)
            ],
        }

    def set_state(self, state: dict[str, Any]) -> None:  # noqa: D102
        self._queue = [(self.runhistory.get_config(id), n) for id, n in state["queue"]]

    def __iter__(self) -> Iterator[TrialInfo]:
        """This iter method holds the logic for the intensification loop.
        Some facts about the loop:

        - Adds existing configurations from the runhistory to the queue (that means it supports user-inputs).
        - Everytime an incumbent (with the lowest amount of trials) is intensified, a new challenger is added to the
          queue.
        - If all incumbents are evaluated on the same trials, a new trial is added to one of the incumbents.
        - Only challengers which are not rejected/running/incumbent are intensified by N*2.

        Returns
        -------
        trials : Iterator[TrialInfo]
            Iterator over the trials.
        """
        self.__post_init__()

        rh = self.runhistory
        assert self._max_config_calls is not None

        # What if there are already trials in the runhistory? Should we queue them up?
        # Because they are part of the runhistory, they might be selected as incumbents. However, they are not
        # intensified because they are not part of the queue. We could add them here to incorporate them in the
        # intensification process.
        # Idea: Add all configs to queue (if it is an incumbent it is removed automatically later on)
        # N=1 is enough here as it will increase automatically in the iterations if the configuration is worthy
        # Note: The incumbents are updated once the runhistory is set (see abstract intensifier)
        # Note 2: If the queue was restored, we don't want to go in here (queue is restored)
        if len(self._queue) == 0:
            for config in rh.get_configs():
                hash = get_config_hash(config)
                self._queue.append((config, 1))
                logger.info(f"Added config {hash} from runhistory to the intensifier queue.")

        fails = -1
        while True:
            fails += 1

            # Some criteria to stop the intensification if nothing can be intensified anymore
            if fails > self._retries:
                logger.error("Intensifier could not find any new trials.")
                return

            # Some configs from the runhistory
            running_configs = rh.get_running_configs()
            rejected_configs = self.get_rejected_configs()

            # Now we get the incumbents sorted by number of trials
            # Also, incorporate ``get_incumbent_instance_seed_budget_keys`` here because challengers are only allowed to
            # sample from the incumbent's instances
            incumbents = self.get_incumbents(sort_by="num_trials")
            incumbent_isb_keys = self.get_incumbent_instance_seed_budget_keys()

            # Check if configs in queue are still running
            all_configs_running = True
            for config, _ in self._queue:
                if config not in running_configs:
                    all_configs_running = False
                    break

            if len(self._queue) == 0 or all_configs_running:
                if len(self._queue) == 0:
                    logger.debug("Queue is empty:")
                else:
                    logger.debug("All configs in the queue are running:")

                if len(incumbents) == 0:
                    logger.debug("--- No incumbent to intensify.")

                for incumbent in incumbents:
                    # Instances of this particular incumbent
                    individual_incumbent_isb_keys = rh.get_instance_seed_budget_keys(incumbent)
                    incumbent_hash = get_config_hash(incumbent)

                    # We don't want to intensify an incumbent which is either still running or rejected
                    if incumbent in running_configs:
                        logger.debug(
                            f"--- Skipping intensifying incumbent {incumbent_hash} because it has trials pending."
                        )
                        continue

                    if incumbent in rejected_configs:
                        # This should actually not happen because if a config is rejected the incumbent should
                        # have changed
                        # However, we just keep it here as sanity check
                        logger.debug(f"--- Skipping intensifying incumbent {incumbent_hash} because it was rejected.")
                        continue

                    # If incumbent was evaluated on all incumbent instance intersections but was not evaluated on
                    # the differences, we have to add it here
                    incumbent_isb_key_differences = self.get_incumbent_instance_seed_budget_key_differences()

                    # We set shuffle to false because we first want to evaluate the incumbent instances, then the
                    # differences (to make the instance-seed keys for the incumbents equal again)
                    trials = self._get_next_trials(
                        incumbent,
                        from_keys=incumbent_isb_keys + incumbent_isb_key_differences,
                        shuffle=False,
                    )

                    # If we don't receive any trials, then we try it randomly with any other because we want to
                    # intensify for sure
                    if len(trials) == 0:
                        logger.debug(
                            f"--- Incumbent {incumbent_hash} was already evaluated on all incumbent instances "
                            "and incumbent instance differences so far. Looking for new instances..."
                        )
                        trials = self._get_next_trials(incumbent)
                        logger.debug(f"--- Randomly found {len(trials)} new trials.")

                    if len(trials) > 0:
                        fails = -1
                        logger.debug(
                            f"--- Yielding trial {len(individual_incumbent_isb_keys)+1} of "
                            f"{self._max_config_calls} from incumbent {incumbent_hash}..."
                        )
                        yield trials[0]
                        logger.debug(f"--- Finished yielding for config {incumbent_hash}.")

                        # We break here because we only want to intensify one more trial of one incumbent
                        break
                    else:
                        # assert len(incumbent_isb_keys) == self._max_config_calls
                        logger.debug(
                            f"--- Skipped intensifying incumbent {incumbent_hash} because no new trials have "
                            "been found. Evaluated "
                            f"{len(individual_incumbent_isb_keys)}/{self._max_config_calls} trials."
                        )

                # For each intensification of the incumbent, we also want to intensify the next configuration
                # We simply add it to the queue and intensify it in the next iteration
                try:
                    config = next(self.config_generator)
                    config_hash = get_config_hash(config)
                    self._queue.append((config, 1))
                    logger.debug(f"--- Added a new config {config_hash} to the queue.")

                    # If we added a new config, then we did something in this iteration
                    fails = -1
                except StopIteration:
                    # We stop if we don't find any configuration anymore
                    logger.warning(
                        "If you assume your configspace was not yet exhausted, try to "
                        "increase the number of retries in the config selector."
                    )
                    return
            else:
                logger.debug("Start finding a new challenger in the queue:")
                for i, (config, N) in enumerate(self._queue.copy()):
                    config_hash = get_config_hash(config)

                    # If the config is still running, we ignore it and head to the next config
                    if config in running_configs:
                        logger.debug(f"--- Config {config_hash} is still running. Skipping this config in the queue...")
                        continue

                    # We want to get rid of configs in the queue which are rejected
                    if config in rejected_configs:
                        logger.debug(f"--- Config {config_hash} was removed from the queue because it was rejected.")
                        self._queue.remove((config, N))
                        continue

                    # We don't want to intensify an incumbent here
                    if config in incumbents:
                        logger.debug(f"--- Config {config_hash} was removed from the queue because it is an incumbent.")
                        self._queue.remove((config, N))
                        continue

                    # And then we yield as many trials as we specified N
                    # However, only the same instances as the incumbents are used
                    isk_keys: list[InstanceSeedBudgetKey] | None = None
                    if len(incumbent_isb_keys) > 0:
                        isk_keys = incumbent_isb_keys

                    # TODO: What to do if there are no incumbent instances? (Use-case: call multiple asks)

                    trials = self._get_next_trials(config, N=N, from_keys=isk_keys)
                    logger.debug(f"--- Yielding {len(trials)} trials to evaluate config {config_hash}...")
                    for trial in trials:
                        fails = -1
                        yield trial

                    logger.debug(f"--- Finished yielding for config {config_hash}.")

                    # Now we have to remove the config
                    self._queue.remove((config, N))
                    logger.debug(f"--- Removed config {config_hash} with N={N} from queue.")

                    # Finally, we add the same config to the queue with a higher N
                    # If the config was rejected by the runhistory, then it's been removed in the next iteration
                    if N < self._max_config_calls:
                        new_pair = (config, N * 2)
                        if new_pair not in self._queue:
                            logger.debug(
                                f"--- Doubled trials of config {config_hash} to N={N*2} and added it to the queue "
                                "again."
                            )
                            self._queue.append((config, N * 2))

                            # Also reset fails here
                            fails = -1
                        else:
                            logger.debug(f"--- Config {config_hash} with N={N*2} is already in the queue.")

                    # If we are at this point, it really is important to break because otherwise, we would intensify
                    # all configs in the queue in one iteration
                    break

    def _get_next_trials(
        self,
        config: Configuration,
        *,
        N: int | None = None,
        from_keys: list[InstanceSeedBudgetKey] | None = None,
        shuffle: bool = True,
    ) -> list[TrialInfo]:
        """Returns the next trials of the configuration based on ``get_trials_of_interest``. If N is specified,
        maximum N trials are returned but not necessarily all of them (depending on evaluated already or still running).

        Parameters
        ----------
        N : int | None, defaults to None
            The maximum number of trials to return. If None, all trials (``max_config_calls``) are returned.
            Running and evaluated trials are counted in.
        from_keys : list[InstanceSeedBudgetKey], defaults to None
            Only instances from the list are considered for the trials.
        shuffle : bool, defaults to True
            Shuffles the trials in groups. First, all instances are shuffled, then all seeds.
        """
        rh = self.runhistory
        is_keys = self.get_instance_seed_keys_of_interest()

        # Create trials from the instance seed pairs
        # trials: list[TrialInfo] = []
        # for is_key in is_keys:
        #    trials.append(TrialInfo(config=config, instance=is_key.instance, seed=is_key.seed))

        # Keep ``from_keys`` trials only
        if from_keys is not None:
            valid_is_keys = [key.get_instance_seed_key() for key in from_keys]
            for is_key in is_keys.copy():
                if is_key not in valid_is_keys:
                    is_keys.remove(is_key)

        # Counter is important to actually subtract the number of trials that are already evaluated/running
        # Otherwise, evaluated/running trials are not considered
        # Example: max_config_calls=16, N=8, 2 trials are running, 2 trials are evaluated, 4 trials are pending
        # Without a counter, we would return 8 trials because there are still so many trials left open
        # With counter, we would return only 4 trials because 4 trials are already evaluated/running
        counter = 0

        # Now we actually have to check whether the trials have been evaluated already
        evaluated_isb_keys = rh.get_instance_seed_budget_keys(config, highest_observed_budget_only=False)
        for isb_key in evaluated_isb_keys:
            is_key = isb_key.get_instance_seed_key()
            if is_key in is_keys:
                counter += 1
                is_keys.remove(is_key)

        # It's also important to remove running trials from the selection (we don't want to queue them again)
        running_trials = rh.get_running_trials(config)
        for trial in running_trials:
            is_key = trial.get_instance_seed_key()
            if is_key in is_keys:
                counter += 1
                is_keys.remove(is_key)

        if shuffle:
            is_keys = self._reorder_instance_seed_keys(is_keys)

        # Return only N trials
        if N is not None:
            N = N - counter
            if len(is_keys) > N:
                is_keys = is_keys[:N]

        # Now we convert to trials
        trials: list[TrialInfo] = []
        for is_key in is_keys:
            trials.append(TrialInfo(config=config, instance=is_key.instance, seed=is_key.seed))

        return trials



from __future__ import annotations

import numpy as np

from smac.intensifier.successive_halving import SuccessiveHalving


def determine_HB(min_budget: float, max_budget: float, eta: int = 3) -> dict:
    """Determine one Hyperband round

    Parameters
    ----------
    min_budget : float
        Minimum budget per trial in fidelity units
    max_budget : float
        Maximum budget per trial in fidelity units
    eta : int, defaults to 3
        Input that controls the proportion of configurations discarded in each round of Successive Halving.

    Returns
    -------
    dict
        Info about the Hyperband round
            "max_iterations"
            "n_configs_in_stage"
            "budgets_in_stage"
            "trials_used"
            "budget_used"
            "number_of_brackets"

    """
    _s_max = SuccessiveHalving._get_max_iterations(eta, max_budget, min_budget)

    _max_iterations: dict[int, int] = {}
    _n_configs_in_stage: dict[int, list] = {}
    _budgets_in_stage: dict[int, list] = {}

    for i in range(_s_max + 1):
        max_iter = _s_max - i

        _budgets_in_stage[i], _n_configs_in_stage[i] = SuccessiveHalving._compute_configs_and_budgets_for_stages(
            eta, max_budget, max_iter, _s_max
        )
        _max_iterations[i] = max_iter + 1

    total_trials = np.sum([np.sum(v) for v in _n_configs_in_stage.values()])
    total_budget = np.sum([np.sum(v) for v in _budgets_in_stage.values()])

    return {
        "max_iterations": _max_iterations,
        "n_configs_in_stage": _n_configs_in_stage,
        "budgets_in_stage": _budgets_in_stage,
        "trials_used": total_trials,
        "budget_used": total_budget,
        "number_of_brackets": len(_max_iterations),
    }


def determine_hyperband_for_multifidelity(
    total_budget: float, min_budget: float, max_budget: float, eta: int = 3
) -> dict:
    """Determine how many Hyperband rounds should happen based on a total budget

    Parameters
    ----------
    total_budget : float
        Total budget for the complete optimization in fidelity units
    min_budget : float
        Minimum budget per trial in fidelity units
    max_budget : float
        Maximum budget per trial in fidelity units
    eta : int, defaults to 3
        Input that controls the proportion of configurations discarded in each round of Successive Halving.

    Returns
    -------
    dict
        Info about one Hyperband round
            "max_iterations"
            "n_configs_in_stage"
            "budgets_in_stage"
            "trials_used"
            "budget_used"
            "number_of_brackets"
        Info about whole optimization
            "n_trials"
            "total_budget"
            "eta"
            "min_budget"
            "max_budget"

    """
    # Determine the HB
    hyperband_round = determine_HB(eta=eta, min_budget=min_budget, max_budget=max_budget)

    # Calculate how many HB rounds we can have
    budget_used_per_hyperband_round = hyperband_round["budget_used"]
    number_of_full_hb_rounds = int(np.floor(total_budget / budget_used_per_hyperband_round))
    remaining_budget = total_budget % budget_used_per_hyperband_round
    trials_used_per_hb_round = hyperband_round["trials_used"]
    n_configs_in_stage = hyperband_round["n_configs_in_stage"]
    budgets_in_stage = hyperband_round["budgets_in_stage"]

    remaining_trials = 0
    for stage in n_configs_in_stage.keys():
        B = budgets_in_stage[stage]
        C = n_configs_in_stage[stage]
        for b, c in zip(B, C):
            # How many trials are left?
            # If b * c is lower than remaining budget, we can add full c
            # otherwise we need to find out how many trials we can do with this budget
            remaining_trials += min(c, int(np.floor(remaining_budget / b)))
            # We cannot go lower than 0
            # If we are in the case of b*c > remaining_budget, we will not have any
            # budget left. We can not add full c but the number of trials that still fit
            remaining_budget = max(0, remaining_budget - b * c)

    n_trials = int(number_of_full_hb_rounds * trials_used_per_hb_round + remaining_trials)

    hyperband_info = hyperband_round
    hyperband_info["n_trials"] = n_trials
    hyperband_info["total_budget"] = total_budget
    hyperband_info["eta"] = eta
    hyperband_info["min_budget"] = min_budget
    hyperband_info["max_budget"] = max_budget

    return hyperband_info


def print_hyperband_summary(hyperband_info: dict) -> None:
    """Print summary about Hyperband as used in the MultiFidelityFacade

    Parameters
    ----------
    hyperband_info : dict
        Info dict about Hyperband
    """
    print("-" * 30, "HYPERBAND IN MULTI-FIDELITY", "-" * 30)
    print("total budget:\t\t", hyperband_info["total_budget"])
    print("total number of trials:\t", hyperband_info["n_trials"])
    print("number of HB rounds:\t", hyperband_info["total_budget"] / hyperband_info["budget_used"])
    print()

    print("\t~~~~~~~~~~~~HYPERBAND ROUND")
    print("\teta:\t\t\t\t\t", hyperband_info["eta"])
    print("\tmin budget per trial:\t\t\t", hyperband_info["min_budget"])
    print("\tmax budget per trial:\t\t\t", hyperband_info["max_budget"])
    print("\ttotal number of trials per HB round:\t", hyperband_info["trials_used"])
    print("\tbudget used per HB round:\t\t", hyperband_info["budget_used"])
    print("\tnumber of brackets:\t\t\t", hyperband_info["number_of_brackets"])
    print("\tbudgets per stage:\t\t\t", hyperband_info["budgets_in_stage"])
    print("\tn configs per stage:\t\t\t", hyperband_info["n_configs_in_stage"])
    print("-" * (2 * 30 + len("HYPERBAND IN MULTI-FIDELITY") + 2))


def get_n_trials_for_hyperband_multifidelity(
    total_budget: float, min_budget: float, max_budget: float, eta: int = 3, print_summary: bool = True
) -> int:
    """Calculate the number of trials needed for multi-fidelity optimization

    Specify the total budget and find out how many trials that equals.

    Parameters
    ----------
    total_budget : float
        Total budget for the complete optimization in fidelity units.
        A fidelity unit can be one epoch or a fraction of a dataset size.
    min_budget : float
        Minimum budget per trial in fidelity units
    max_budget : float
        Maximum budget per trial in fidelity units
    eta : int, defaults to 3
        Input that controls the proportion of configurations discarded in each round of Successive Halving.

    Returns
    -------
    int
        Number of trials needed for the specified total budgets
    """
    hyperband_info = determine_hyperband_for_multifidelity(
        total_budget=total_budget, eta=eta, min_budget=min_budget, max_budget=max_budget
    )
    if print_summary:
        print_hyperband_summary(hyperband_info=hyperband_info)
    return hyperband_info["n_trials"]



from __future__ import annotations

from abc import abstractmethod

from src.datasets.dataset import Dataset
from src.models.model import Model
from src.tasks.task import Task


class Wrapper:
    supported_versions: list[str] = []

    def __init__(self, task: Task, seed: int = 0) -> None:
        if len(task.objectives) > 2:
            raise RuntimeError("Not supported yet.")

        self._seed = seed
        self._task = task
        
    @property
    def seed(self) -> int:
        return self._seed

    @property
    def task(self) -> Task:
        return self._task

    @property
    def model(self) -> Model:
        return self._task.model

    @property
    def dataset(self) -> Dataset:
        return self._task.model.dataset

    @abstractmethod
    def run(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_trajectory(self, sort_by: str = "trials") -> tuple[list[float], list[float]]:
        """List of x and y values of the incumbents over time. x depends on ``sort_by``."""
        raise NotImplementedError

    @abstractmethod
    def get_samples(self) -> tuple[list["Configuration"]]:
        raise NotImplementedError



from __future__ import annotations

from typing import Any

import numpy as np
from scipy.stats import norm

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class PI(AbstractAcquisitionFunction):
    r"""Probability of Improvement

    :math:`P(f_{t+1}(\mathbf{X})\geq f(\mathbf{X^+}))` :math:`:= \Phi(\\frac{ \mu(\mathbf{X})-f(\mathbf{X^+}) }
    { \sigma(\mathbf{X}) })` with :math:`f(X^+)` as the incumbent and :math:`\Phi` the cdf of the standard normal.

    Parameters
    ----------
    xi : float, defaults to 0.0
        Controls the balance between exploration and exploitation of the acquisition function.
    """

    def __init__(self, xi: float = 0.0):
        super(PI, self).__init__()
        self._xi: float = xi
        self._eta: float | None = None

    @property
    def name(self) -> str:  # noqa: D102
        return "Probability of Improvement"

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"xi": self._xi})

        return meta

    def _update(self, **kwargs: Any) -> None:
        """Update acsquisition function attributes

        Parameters
        ----------
        eta : float
            Function value of current incumbent.
        xi : float, optional
            Exploration-exploitation trade-off parameter
        """
        assert "eta" in kwargs
        self._eta = kwargs["eta"]

        if "xi" in kwargs and kwargs["xi"] is not None:
            self._xi = kwargs["xi"]

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Compute the PI value.

        Parameters
        ----------
        X: np.ndarray [N, D]
           Points to evaluate PI. N is the number of points and D the dimension for the points.

        Returns
        -------
        np.ndarray [N, 1]
            Expected Improvement of X.

        Raises
        ------
        ValueError
            If `update` has not been called before (current incumbent value `eta` unspecified).

        """
        assert self._model is not None
        if self._eta is None:
            raise ValueError(
                "No current best specified. Call update("
                "eta=<float>) to inform the acquisition function "
                "about the current best value."
            )

        if len(X.shape) == 1:
            X = X[:, np.newaxis]
        m, var_ = self._model.predict_marginalized(X)
        std = np.sqrt(var_)

        return norm.cdf((self._eta - m - self._xi) / std)



from smac.initial_design.abstract_initial_design import AbstractInitialDesign
from smac.initial_design.default_design import DefaultInitialDesign
from smac.initial_design.factorial_design import FactorialInitialDesign
from smac.initial_design.latin_hypercube_design import LatinHypercubeInitialDesign
from smac.initial_design.random_design import RandomInitialDesign
from smac.initial_design.sobol_design import SobolInitialDesign

__all__ = [
    "AbstractInitialDesign",
    "LatinHypercubeInitialDesign",
    "FactorialInitialDesign",
    "RandomInitialDesign",
    "SobolInitialDesign",
    "DefaultInitialDesign",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

from inspect import Signature, signature

import numpy as np
import sklearn.gaussian_process.kernels as kernels

from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior
from smac.utils.configspace import get_conditional_hyperparameters

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class AbstractKernel:
    """
    This is a mixin for a kernel to override functions of the kernel. Because it overrides functions of the kernel,
    it needs to be placed first in the inheritance hierarchy. For this reason it is not possible to subclass the
    Mixin from the kernel class because this will prevent it from being instantiatable. Therefore, mypy won't know about
    anything related to the superclass and some type:ignore statements has to be added when accessing a member that is
    declared in the superclass such as `self.has_conditions`, `self._call`, `super().get_params`, etc.

    Parameters
    ----------
    operate_on : np.ndarray, defaults to None
        On which numpy array should be operated on.
    has_conditions : bool, defaults to False
        Whether the kernel has conditions.
    prior : AbstractPrior, defaults to None
        Which prior the kernel is using.

    Attributes
    ----------
    operate_on : np.ndarray, defaults to None
        On which numpy array should be operated on.
    has_conditions : bool, defaults to False
        Whether the kernel has conditions. Might be changed by the gaussian process.
    prior : AbstractPrior, defaults to None
        Which prior the kernel is using. Primarily used by sklearn.
    """

    def __init__(
        self,
        *,
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
        prior: AbstractPrior | None = None,
        **kwargs: Any,
    ) -> None:
        self.operate_on = operate_on
        self.has_conditions = has_conditions
        self.prior = prior
        self._set_active_dims(operate_on)

        # Since this class is a mixin, we just pass all the other parameters to the next class.
        super().__init__(**kwargs)

        # Get variables from next class:
        # We make it explicit here to make sure the next class really has this attributes.
        self._hyperparameters: list[kernels.Hyperparameter] = super().hyperparameters  # type: ignore
        self._n_dims: int = super().n_dims  # type: ignore
        self._len_active: int | None

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object. This method calls the `get_params` method to collect the
        parameters of the kernel.
        """
        meta: dict[str, Any] = {"name": self.__class__.__name__}
        meta.update(self.get_params(deep=False))

        # We have to handle some special cases to make the meta data serializable
        for k in meta:
            v = meta[k]
            if isinstance(v, AbstractKernel):
                meta[k] = v.meta

            if isinstance(v, AbstractPrior):
                meta[k] = v.meta

            if isinstance(v, np.ndarray):
                meta[k] = v.tolist()

        return meta

    @property
    def hyperparameters(self) -> list[kernels.Hyperparameter]:
        """Returns a list of all hyperparameter specifications."""
        return self._hyperparameters

    @property
    def n_dims(self) -> int:
        """Returns the number of non-fixed hyperparameters of the kernel."""
        return self._n_dims

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of this kernel.

        Parameters
        ----------
        deep : bool, defaults to True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict[str, Any]
            Parameter names mapped to their values.
        """
        params = {}

        # ignore[misc] looks like it catches all kinds of errors, but misc is actually a category from mypy:
        # https://mypy.readthedocs.io/en/latest/error_code_list.html#miscellaneous-checks-misc
        tmp = super().get_params(deep)  # type: ignore[misc] # noqa F821
        args = list(tmp.keys())

        # Sum and Product do not clone the 'has_conditions' attribute by default. Instead of changing their
        # get_params() method, we simply add the attribute here!
        if "has_conditions" not in args:
            args.append("has_conditions")

        for arg in args:
            params[arg] = getattr(self, arg, None)

        return params

    def __call__(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Call the kernel function. Internally, `self._call` is called, which must be specified by a subclass."""
        if active is None and self.has_conditions:
            if self.operate_on is None:
                active = get_conditional_hyperparameters(X, Y)
            else:
                if Y is None:
                    active = get_conditional_hyperparameters(X[:, self.operate_on], None)
                else:
                    active = get_conditional_hyperparameters(X[:, self.operate_on], Y[:, self.operate_on])

        if self.operate_on is None:
            rval = self._call(X, Y, eval_gradient, active)
        else:
            if self._len_active is None:
                raise RuntimeError("The internal variable `_len_active` is not set.")

            if Y is None:
                rval = self._call(
                    X=X[:, self.operate_on].reshape([-1, self._len_active]),
                    Y=None,
                    eval_gradient=eval_gradient,
                    active=active,
                )
                X = X[:, self.operate_on].reshape((-1, self._len_active))
            else:
                rval = self._call(
                    X=X[:, self.operate_on].reshape([-1, self._len_active]),
                    Y=Y[:, self.operate_on].reshape([-1, self._len_active]),
                    eval_gradient=eval_gradient,
                    active=active,
                )
                X = X[:, self.operate_on].reshape((-1, self._len_active))
                Y = Y[:, self.operate_on].reshape((-1, self._len_active))

        return rval

    def __add__(self, b: kernels.Kernel | float) -> kernels.Sum:
        if not isinstance(b, kernels.Kernel):
            return SumKernel(self, ConstantKernel(b))

        return SumKernel(self, b)

    def __radd__(self, b: kernels.Kernel | float) -> kernels.Sum:
        if not isinstance(b, kernels.Kernel):
            return SumKernel(ConstantKernel(b), self)

        return SumKernel(b, self)

    def __mul__(self, b: kernels.Kernel | float) -> kernels.Product:
        if not isinstance(b, kernels.Kernel):
            return ProductKernel(self, ConstantKernel(b))

        return ProductKernel(self, b)

    def __rmul__(self, b: kernels.Kernel | float) -> kernels.Product:
        if not isinstance(b, kernels.Kernel):
            return ProductKernel(ConstantKernel(b), self)

        return ProductKernel(b, self)

    @abstractmethod
    def _call(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Return the kernel k(X, Y) and optionally its gradient.

        Note
        ----
        Code partially copied from skopt (https://github.com/scikit-optimize).
        Made small changes to only compute necessary values and use scikit-learn helper functions.

        Parameters
        ----------
        X : np.ndarray [#samples, #features]
            Left argument of the returned kernel k(X, Y).
        Y : np.ndarray [#samples, #features], defaults to None
            Right argument of the returned kernel k(X, Y). If None, k(X, X) is evaluated instead.
        eval_gradient : bool, defaults to False
            Determines whether the gradient with respect to the kernel hyperparameter is determined.
            Only supported when `Y` is None.
        active : np.ndarray [#samples, #features], defaults to None
            Boolean array specifying which hyperparameters are active.

        Returns
        -------
        K : np.ndarray [#X_samples, #Y_samples]
            Kernel k(X, Y).
        K_gradient : np.ndarray [#X_samples, #X_samples, #dimensions]
            The gradient of the kernel k(X, X) with respect to the hyperparameter of the kernel.
            Only returned when `eval_gradient` is True.
        """
        raise NotImplementedError

    def _signature(self, func: Callable) -> Signature:
        sig_: Signature | None

        try:
            sig_ = self._signature_cache.get(func)
        except AttributeError:
            self._signature_cache: dict[Callable, Signature] = {}
            sig_ = None

        if sig_ is None:
            sig = signature(func)
            self._signature_cache[func] = sig

            return sig
        else:
            return sig_

    def _set_active_dims(self, operate_on: np.ndarray | None = None) -> None:
        """Sets dimensions this kernel should work on."""
        if operate_on is not None and isinstance(operate_on, (list, np.ndarray)):
            if not isinstance(operate_on, np.ndarray):
                raise TypeError(f"The argument `operate_on` needs to be of type np.ndarray but is {type(operate_on)}")

            if not np.issubdtype(operate_on.dtype, np.integer):
                raise ValueError(f"The dtype of `operate_on` needs to be np.integer, but is {operate_on.dtype}")

            self.operate_on = operate_on
            self._len_active = len(operate_on)
        else:
            self.operate_on = None
            self._len_active = None


class SumKernel(AbstractKernel, kernels.Sum):
    """Sum kernel implementation."""

    def __init__(
        self,
        k1: kernels.Kernel,
        k2: kernels.Kernel,
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
    ) -> None:
        super().__init__(
            operate_on=operate_on,
            has_conditions=has_conditions,
            k1=k1,
            k2=k2,
        )

    def __call__(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y).

        Y : np.ndarray, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined.

        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.

        Returns
        -------
        K : np.ndarray, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y).

        K_gradient : np.ndarray (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(X, Y, eval_gradient=True, active=active)
            K2, K2_gradient = self.k2(X, Y, eval_gradient=True, active=active)

            return K1 + K2, np.dstack((K1_gradient, K2_gradient))
        else:
            return self.k1(X, Y, active=active) + self.k2(X, Y, active=active)


class ProductKernel(AbstractKernel, kernels.Product):
    """Product kernel implementation."""

    def __init__(
        self,
        k1: kernels.Kernel,
        k2: kernels.Kernel,
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
    ) -> None:
        super().__init__(
            operate_on=operate_on,
            has_conditions=has_conditions,
            k1=k1,
            k2=k2,
        )

    def __call__(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y).

        Y : np.ndarray, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined.

        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.

        Returns
        -------
        K : np.ndarray, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y).

        K_gradient : np.ndarray (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(X, Y, eval_gradient=True, active=active)
            K2, K2_gradient = self.k2(X, Y, eval_gradient=True, active=active)

            return K1 * K2, np.dstack((K1_gradient * K2[:, :, np.newaxis], K2_gradient * K1[:, :, np.newaxis]))
        else:
            return self.k1(X, Y, active=active) * self.k2(X, Y, active=active)


class ConstantKernel(AbstractKernel, kernels.ConstantKernel):
    def __init__(
        self,
        constant_value: float = 1.0,
        constant_value_bounds: tuple[float, float] = (1e-5, 1e5),
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
        prior: AbstractPrior | None = None,
    ) -> None:
        super().__init__(
            operate_on=operate_on,
            has_conditions=has_conditions,
            prior=prior,
            constant_value=constant_value,
            constant_value_bounds=constant_value_bounds,
        )

    def __call__(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y).

        Y : np.ndarray, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            is evaluated instead.

        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.

        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.

        Returns
        -------
        K : np.ndarray, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y).

        K_gradient : np.ndarray (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        X = np.atleast_2d(X)
        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        K = np.full(
            (X.shape[0], Y.shape[0]),
            self.constant_value,
            dtype=np.array(self.constant_value).dtype,
        )
        if eval_gradient:
            if not self.hyperparameter_constant_value.fixed:
                return (
                    K,
                    np.full(
                        (X.shape[0], X.shape[0], 1),
                        self.constant_value,
                        dtype=np.array(self.constant_value).dtype,
                    ),
                )
            else:
                return K, np.empty((X.shape[0], X.shape[0], 0))
        else:
            return K



from __future__ import annotations

from typing import TypeVar

import numpy as np

from smac.model.abstract_model import AbstractModel

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


Self = TypeVar("Self", bound="MultiObjectiveModel")


class MultiObjectiveModel(AbstractModel):
    """Wrapper for the surrogate model to predict multiple objectives.

    Parameters
    ----------
    models : AbstractModel | list[AbstractModel]
        Which model should be used. If it is a list, then it must provide as many models as objectives.
        If it is a single model only, the model is used for all objectives.
    objectives : list[str]
        Which objectives should be used.
    seed : int
    """

    def __init__(
        self,
        models: AbstractModel | list[AbstractModel],
        objectives: list[str],
        seed: int = 0,
    ) -> None:
        self._n_objectives = len(objectives)
        if isinstance(models, list):
            assert len(models) == len(objectives)

            # Make sure the configspace is the same
            configspace = models[0]._configspace
            for m in models:
                assert configspace == m._configspace

            self._models = models
        else:
            configspace = models._configspace
            self._models = [models for _ in range(self._n_objectives)]

        super().__init__(
            configspace=configspace,
            instance_features=None,
            pca_components=None,
            seed=seed,
        )

    @property
    def models(self) -> list[AbstractModel]:
        """The internally used surrogate models."""
        return self._models

    def predict_marginalized(self, X: np.ndarray) -> tuple[np.ndarray, np.ndarray]:  # noqa: D102
        mean = np.zeros((X.shape[0], self._n_objectives))
        var = np.zeros((X.shape[0], self._n_objectives))

        for i, estimator in enumerate(self._models):
            m, v = estimator.predict_marginalized(X)
            mean[:, i] = m.flatten()
            var[:, i] = v.flatten()

        return mean, var

    def _train(self: Self, X: np.ndarray, Y: np.ndarray) -> Self:
        if len(self._models) == 0:
            raise ValueError("The list of surrogate models is empty.")

        for i, model in enumerate(self._models):
            model.train(X, Y[:, i])

        return self

    def _predict(
        self,
        X: np.ndarray,
        covariance_type: str | None = "diagonal",
    ) -> tuple[np.ndarray, np.ndarray | None]:
        if covariance_type != "diagonal":
            raise ValueError("`covariance_type` can only take `diagonal` for this model.")

        mean = np.zeros((X.shape[0], self._n_objectives))
        var = np.zeros((X.shape[0], self._n_objectives))

        for i, estimator in enumerate(self._models):
            m, v = estimator.predict(X)
            assert v is not None
            mean[:, i] = m.flatten()
            var[:, i] = v.flatten()

        return mean, var



from __future__ import annotations

from enum import IntEnum

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class StatusType(IntEnum):
    """Class to define status types of configs."""

    RUNNING = 0  # In case a job was submitted, but it has not finished.
    SUCCESS = 1
    CRASHED = 2
    TIMEOUT = 3
    MEMORYOUT = 4



from __future__ import annotations

from typing import Iterable

from smac.utils.logging import get_logger

logger = get_logger(__name__)


def recursively_compare_dicts(
    d1: dict,
    d2: dict,
    *,
    level: str = "root",
    diff: list[str] | None = None,
) -> list[str]:
    """Compares dictionaries recursively. Returns a list of differences in string format.

    Parameters
    ----------
    d1 : dict
        First dictionary.
    d2 : dict
        Second dictionary.
    level : str, defaults to "root"
        How the first level is called.
    diff : list[str] | None, defaults to None
        Used for recursion.

    Returns
    -------
    list[str]
        List of differences in string format.
    """
    if diff is None:
        diff = []

    if isinstance(d1, dict) and isinstance(d2, dict):
        if d1.keys() != d2.keys():
            s1 = set(d1.keys())
            s2 = set(d2.keys())
            # logger.info("{:<20} + {} - {}".format(level, s1 - s2, s2 - s1))
            # logger.info("{} - {}".format(s1 - s2, s2 - s1))
            diff += [f"{level} + {s1 - s2} - {s2 - s1}"]
            common_keys = s1 & s2
        else:
            common_keys = set(d1.keys())

        for k in common_keys:
            recursively_compare_dicts(d1[k], d2[k], level="{}.{}".format(level, k), diff=diff)

    elif isinstance(d1, list) and isinstance(d2, list):
        if len(d1) != len(d2):
            diff += [f"{level}: len1={len(d1)}; len2={len(d2)}"]
            # logger.info("{:<20} len1={}; len2={}".format(level, len(d1), len(d2)))
            # logger.info("len1={}; len2={}".format(len(d1), len(d2)))
        common_len = min(len(d1), len(d2))

        for i in range(common_len):
            recursively_compare_dicts(d1[i], d2[i], level="{}[{}]".format(level, i), diff=diff)

    else:
        if d1 != d2:
            diff += [f"{level}: {d1} != {d2}"]
            # logger.info("{:<20} {} != {}".format(level, d1, d2))
            # logger.info("len1={}; len2={}".format(len(d1), len(d2)))

    return diff


def batch(iterable: list, n: int = 1) -> Iterable[list]:
    """Batches an iterable into chunks of size n."""
    length = len(iterable)
    for ndx in range(0, length, n):
        yield iterable[ndx : min(ndx + n, length)]



from __future__ import annotations

import logging
import logging.config
from pathlib import Path

import yaml
from typing_extensions import Literal

import smac

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


def setup_logging(
    level: int | Path | Literal[False] | None = False,
) -> None:
    """Sets up the logging configuration for all modules.

    Parameters
    ----------
    level : int | Path | Literal[False] | None, defaults to None
        An integer representing the logging level. An custom logging configuration can be used when passing a path.
        If False, no logging setup is performed.
    """
    if level is False:
        return

    if isinstance(level, Path):
        log_filename = level
    else:
        path = Path() / smac.__file__
        log_filename = path.parent / "logging.yml"

    with (log_filename).open("r") as stream:
        config = yaml.safe_load(stream)

    if isinstance(level, int):
        config["root"]["level"] = level
        config["handlers"]["console"]["level"] = level

    logging.config.dictConfig(config)


def get_logger(logger_name: str) -> logging.Logger:
    """Get the logger by name."""
    logger = logging.getLogger(logger_name)
    return logger



from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.function.confidence_bound import LCB
from smac.acquisition.function.expected_improvement import EI, EIPS
from smac.acquisition.function.integrated_acquisition_function import (
    IntegratedAcquisitionFunction,
)
from smac.acquisition.function.prior_acquisition_function import (
    PriorAcquisitionFunction,
)
from smac.acquisition.function.probability_improvement import PI
from smac.acquisition.function.thompson import TS

__all__ = [
    "AbstractAcquisitionFunction",
    "LCB",
    "PI",
    "EI",
    "EIPS",
    "TS",
    "PriorAcquisitionFunction",
    "IntegratedAcquisitionFunction",
]



from __future__ import annotations

from typing import Any

import numpy as np

from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class ModulusRandomDesign(AbstractRandomDesign):
    """Interleave a random configuration after a constant number of configurations found by
    Bayesian optimization.

    Parameters
    ----------
    modulus : float
        Every modulus-th configuration will be at random.
    seed : int
        Integer used to initialize random state. This class does not use the seed.
    """

    def __init__(self, modulus: float = 2.0, seed: int = 0):
        super().__init__(seed)
        assert modulus > 0
        if modulus <= 1.0:
            logger.warning("Using SMAC with random configurations only. ROAR is the better choice for this.")

        self._modulus = modulus

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"modulus": self._modulus})

        return meta

    def check(self, iteration: int) -> bool:  # noqa: D102
        assert iteration >= 0
        return iteration % self._modulus < 1


class DynamicModulusRandomDesign(AbstractRandomDesign):
    """Interleave a random configuration, decreasing the fraction of random configurations over time.

    Parameters
    ----------
    start_modulus : float, defaults to 2.0
       Initially, every modulus-th configuration will be at random.
    modulus_increment : float, defaults to 0.3
       Increase modulus by this amount in every iteration.
    end_modulus : float, defaults to np.inf
       The maximum modulus ever used. If the value is reached before the optimization
       is over, it is not further increased. If it is not reached before the optimization is over,
       there will be no adjustment to make sure that the `end_modulus` is reached.
    seed : int, defaults to 0
        Integer used to initialize the random state. This class does not use the seed.
    """

    def __init__(
        self, start_modulus: float = 2.0, modulus_increment: float = 0.3, end_modulus: float = np.inf, seed: int = 0
    ):
        super().__init__(seed)
        assert start_modulus > 0
        assert modulus_increment > 0
        assert end_modulus > 0
        assert end_modulus > start_modulus

        if start_modulus <= 1.0 and modulus_increment <= 0.0:
            logger.warning("Using SMAC with random configurations only. ROAR is the better choice for this.")

        self._modulus = start_modulus
        self._start_modulus = start_modulus
        self._modulus_increment = modulus_increment
        self._end_modulus = end_modulus

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "start_modulus": self._start_modulus,
                "end_modulus": self._end_modulus,
                "modulus_increment": self._modulus_increment,
            }
        )

        return meta

    def next_iteration(self) -> None:  # noqa: D102
        self._modulus += self._modulus_increment
        self._modulus = min(self._modulus, self._end_modulus)

    def check(self, iteration: int) -> bool:  # noqa: D102
        assert iteration >= 0

        if iteration % self._modulus < 1:
            return True
        else:
            return False
